# qtmui/material/chart/chart_pie.py
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union
from PySide6.QtCore import Qt, QPointF, QRectF
from PySide6.QtWidgets import QWidget, QVBoxLayout, QGraphicsItem
from PySide6.QtCharts import QChart, QChartView, QPieSeries, QPieSlice
from PySide6.QtGui import (
    QPainter, QColor, QPen, QBrush, QFont, QFontMetrics,
    QLinearGradient, QGradient
)

from qtmui.material.styles import useTheme


class TooltipItem(QGraphicsItem):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.lines = []
        self.padding = 10
        self.bg_color = QColor(20, 20, 20, 240)
        self.border_color = QColor(255, 255, 255, 80)
        self.border_radius = 8
        self.font = QFont("Segoe UI", 9)
        self._w = self._h = 0
        self.setZValue(99999)
        self.hide()

    def clear(self):
        self.lines = []
        self._w = self._h = 0
        self.hide()

    def add_line(self, text: str, color: QColor):
        self.lines.append((text, color))
        fm = QFontMetrics(self.font)
        w = fm.horizontalAdvance(text) + 20
        h = fm.height() + 4
        self._w = max(self._w, w)
        self._h += h

    def boundingRect(self) -> QRectF:
        return QRectF(0, 0, self._w + self.padding * 2, self._h + self.padding * 2)

    def paint(self, painter: QPainter, *args):
        if not self.lines:
            return
        painter.setRenderHint(QPainter.Antialiasing)
        painter.setFont(self.font)
        fm = QFontMetrics(self.font)

        rect = self.boundingRect()
        painter.setBrush(QBrush(self.bg_color))
        painter.setPen(QPen(self.border_color, 1))
        painter.drawRoundedRect(rect, self.border_radius, self.border_radius)

        y = self.padding
        for text, color in self.lines:
            painter.setBrush(QBrush(color))
            painter.setPen(Qt.NoPen)
            painter.drawEllipse(self.padding, y + 2, 8, 8)
            painter.setPen(QPen(QColor("white")))
            painter.drawText(self.padding + 16, y + fm.ascent() + 2, text)
            y += fm.height() + 4


class SmartChartView(QChartView):
    def __init__(self, chart, parent=None):
        super().__init__(chart, parent)
        self.owner: Optional['ChartPie'] = None
        self.setMouseTracking(True)

    def mouseMoveEvent(self, event):
        if self.owner and self.owner.current_hover_slice:
            # DÙNG event.pos() → CHÍNH XÁC TUYỆT ĐỐI, CẬP NHẬT MỖI FRAME
            self.owner._update_tooltip_position(event.pos())
        super().mouseMoveEvent(event)

    def leaveEvent(self, event):
        if self.owner:
            self.owner.force_hide_tooltip()
        super().leaveEvent(event)


class ChartPie(QWidget):
    def __init__(
        self,
        dir: str = "ltr",
        type: str = "pie",
        series: Optional[Union[List[Dict[str, Any]], List[float], List[int]]] = None,
        width: Optional[Union[str, int]] = None,
        height: Optional[Union[str, int]] = None,
        options: Optional[Dict[str, Any]] = None,
        key: Optional[str] = None,
        *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.setLayout(QVBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

        self._type = type.lower()
        self._raw_series = series or []
        self._options = options or {}
        self._width = width
        self._height = height

        self._series = self._normalize_series(self._raw_series)
        self.current_hover_slice: Optional[QPieSlice] = None

        self._init_chart()

        theme = useTheme()
        theme.state.valueChanged.connect(self._apply_theme)
        self._apply_theme()

    def _normalize_series(self, raw) -> List[Dict[str, Any]]:
        if not raw:
            return []
        if isinstance(raw, (list, tuple)) and all(isinstance(x, (int, float)) for x in raw):
            labels = self._options.get("labels", [f"Item {i+1}" for i in range(len(raw))])
            return [{"label": labels[i] if i < len(labels) else f"Item {i+1}", "value": float(v)}
                    for i, v in enumerate(raw)]
        if isinstance(raw, (list, tuple)) and all(isinstance(x, dict) for x in raw):
            return [{"label": x.get("label", f"Item {i+1}"), "value": float(x.get("value", 0))}
                    for i, x in enumerate(raw)]
        raise ValueError("series must be List[int/float] or List[dict(label=..., value=...)]")

    def _init_chart(self):
        if self._height: self.setFixedHeight(self._height)
        if self._width:  self.setFixedWidth(self._width)

        self.pie_series = QPieSeries()
        self.pie_series.setHoleSize(0.4 if self._type == "donut" else 0.0)

        total = sum(item["value"] for item in self._series)
        self.colors = []
        palette_names = ["primary", "secondary", "success", "warning", "error", "info"]

        for idx, item in enumerate(self._series):
            label = item["label"]
            value = item["value"]
            percentage = (value / total * 100) if total > 0 else 0

            slice = QPieSlice(label, value)
            slice.setLabel(f"{percentage:.1f}%")
            slice.setLabelVisible(self._options.get("dataLabels", {}).get("enabled", True))
            slice.setLabelPosition(QPieSlice.LabelInsideHorizontal)

            color_key = palette_names[idx % len(palette_names)]
            color = getattr(useTheme().palette, color_key).main
            qcolor = QColor(color)
            self.colors.append(qcolor)
            slice.setBrush(QBrush(qcolor))

            if not self._options.get("stroke", {}).get("show", True):
                slice.setPen(Qt.NoPen)
            else:
                pen = QPen(QColor("#ffffff"))
                pen.setWidth(2)
                slice.setPen(pen)

            slice.hovered.connect(self._on_slice_hovered)
            self.pie_series.append(slice)

        self.chart = QChart()
        self.chart.addSeries(self.pie_series)
        self.chart.setAnimationOptions(QChart.NoAnimation)
        self.chart.setBackgroundRoundness(12)

        legend_opt = self._options.get("legend", {})
        legend = self.chart.legend()
        legend.setVisible(legend_opt.get("show", True) != False)
        pos = legend_opt.get("position", "bottom").lower()
        align_map = {"top": Qt.AlignTop, "bottom": Qt.AlignBottom, "left": Qt.AlignLeft, "right": Qt.AlignRight}
        legend.setAlignment(align_map.get(pos, Qt.AlignBottom))

        self.tooltip = TooltipItem()
        self.tooltip.hide()

        self.chart_view = SmartChartView(self.chart)
        self.chart_view.owner = self
        self.chart_view.setRenderHint(QPainter.Antialiasing)
        self.chart_view.scene().addItem(self.tooltip)
        self.layout().addWidget(self.chart_view)

    def _on_slice_hovered(self, status: bool):
        slice: QPieSlice = self.sender()

        if status:
            self.current_hover_slice = slice
            self._show_tooltip_for_current_slice()
        else:
            if self.current_hover_slice == slice:
                self.current_hover_slice = None
                self.hide_tooltip()

    def _show_tooltip_for_current_slice(self):
        if not self.current_hover_slice:
            return

        slice = self.current_hover_slice
        index = self.pie_series.slices().index(slice)
        item = self._series[index]
        value = item["value"]
        total = sum(s["value"] for s in self._series)
        percentage = (value / total * 100) if total > 0 else 0

        self.tooltip.clear()
        self.tooltip.add_line(item["label"], self.colors[index])
        self.tooltip.add_line(f"{value:,.0f} ({percentage:.1f}%)", QColor("#cccccc"))

        self.tooltip.show()
        # Không cần setPos ở đây → để mouseMoveEvent lo

    def _update_tooltip_position(self, pos: QPointF):
        """Cập nhật vị trí tooltip theo vị trí chuột từ event → MƯỢT NHƯ BƠ"""
        if not self.tooltip.isVisible():
            return

        rect = self.tooltip.boundingRect()
        tx = pos.x() - rect.width() / 2
        ty = pos.y() - rect.height() - 10  # trên đầu chuột 10px

        self.tooltip.setPos(tx, ty)

    def force_hide_tooltip(self):
        self.current_hover_slice = None
        self.tooltip.hide()
        self.tooltip.clear()

    def hide_tooltip(self):
        self.tooltip.hide()
        self.tooltip.clear()

    def _apply_theme(self):
        theme = useTheme()

        grad = QLinearGradient(0, 0, 0, 1)
        grad.setCoordinateMode(QGradient.ObjectBoundingMode)
        grad.setColorAt(0.0, QColor(theme.palette.background.paper))
        grad.setColorAt(1.0, QColor(theme.palette.background.main))
        self.chart.setBackgroundBrush(QBrush(grad))

        text_color = QColor(theme.palette.text.secondary)
        self.chart.setTitleBrush(text_color)
        self.chart.legend().setLabelBrush(text_color)

        palette_names = ["primary", "secondary", "success", "warning", "error", "info"]
        for i, slice in enumerate(self.pie_series.slices()):
            color_key = palette_names[i % len(palette_names)]
            color = getattr(theme.palette, color_key).main
            slice.setBrush(QBrush(QColor(color)))
            self.colors[i] = QColor(color)

        self.chart_view.update()


from typing import Callable
from PySide6.QtCore import QSize, QEvent, Signal, QRect
from PySide6.QtGui import QIcon,  QColor, QBrush
from PySide6.QtWidgets import QStyledItemDelegate, QStyle,  QStyleOptionButton

from .icon import icon_firefox_disabled

from ..utils.icon import icon_base64_to_pixmap


class StyledItemDelegate(QStyledItemDelegate):

    mousePressedSignal = Signal(str)

    def __init__(self, 
                 parent=None,
                 column: int = None,
                 styledOption: object = None,
                 styledOptions: list = None,
                 onMousePressed: Callable = None
                 ):
        super(StyledItemDelegate, self).__init__(parent)
        if onMousePressed is not None:
            self.mousePressedSignal.connect(onMousePressed)
        self.column = column
        self.styledOption = styledOption
        self.styledOptions = styledOptions


    def split_rect(self, rect, num_divisions):
        # Tính toán chiều rộng của mỗi phần chia
        width_per_division = rect.width() // num_divisions
        
        # Tạo danh sách chứa các hình chữ nhật con
        sub_rects = []
        
        # Tạo các hình chữ nhật con
        for i in range(num_divisions):
            if i == 0:
                sub_rect = QRect((rect.left() + i * width_per_division) - 1, rect.top(), width_per_division + 1, rect.height() - 4)
                sub_rects.append(sub_rect)
            if i == 1:
                sub_rect = QRect(rect.left() + i * width_per_division, rect.top(), width_per_division + 1, rect.height() - 4)
                sub_rects.append(sub_rect)
            if i == 2:
                sub_rect = QRect((rect.left() + i * width_per_division) + 1, rect.top(), width_per_division + 1, rect.height() - 4)
                sub_rects.append(sub_rect)
        return sub_rects

    def point_in_rects(self, point, rects):
        for i, rect in enumerate(rects):
            if rect.contains(point):
                return i
        return -1

    def editorEvent(self, event, model, option, index):
        sub_rects = self.split_rect(option.rect, 3)
        test_point = event.pos()

        # print('sub_rects___', sub_rects)

        # Kiểm tra xem test_point có nằm trong hình chữ nhật nào không
        result = self.point_in_rects(test_point, sub_rects)

        if option.state & QStyle.State_MouseOver:
            option.backgroundBrush = QBrush(QColor(211, 211, 211))  # LightGray color

        if index.column() == 10 and event.type() == QEvent.MouseButtonRelease:
            if result != -1:
                if result == 0:
                    self.mousePressedSignal.emit(f"openclose_{index.row()}_{index.column()}")
                if result == 1:
                    self.mousePressedSignal.emit(f"edit_{index.row()}_{index.column()}")
                if result == 2:
                    self.mousePressedSignal.emit(f"delete_{index.row()}_{index.column()}")
                return True
            else:
                print("The point is not in any sub-rectangle.")

        return super().editorEvent(event, model, option, index)

    def paint(self, painter, option, index):
        if index.column() == self.column:  # Assuming the button column is the third column

            styledOption = QStyleOptionButton()
            styledOption.option.rect = option.rect

            styledOption.rect = option.rect
            styledOption.option.icon = QIcon()
            styledOption.option.iconSize = QSize(15, 15)
            styledOption.option.state |= QStyle.State_Enabled
            styledOption.option.icon.addPixmap(icon_base64_to_pixmap(icon_firefox_disabled), QIcon.Normal, QIcon.Off)
            styledOption.option.text = "Open"
            
            painter.save()
            styledOption.style().drawControl(QStyle.CE_PushButton, styledOption.option, painter, styledOption)
            # QApplication.style().drawControl(QStyle.CE_PushButton, opt, painter, self.button)
            painter.restore()

        super(StyledItemDelegate, self).paint(painter, option, index)


from .form_provider import FormProvider
from .rhf_autocomplete import RHFAutocomplete
from .rhf_checkbox import RHFCheckbox, RHFMultiCheckbox
from .rhf_radio_group import RHFRadioGroup
from .rhf_select import RHFSelect, RHFMultiSelect
from .rhf_slider import RHFSlider
from .rhf_switch import RHFSwitch
from .rhf_text_field import RHFTextField
from .rhf_submit import SubmitButton
from .rhf_upload import RHFUpload, RHFUploadBox, RHFUploadAvatar
from .rhf_rating import RHFRating
from .rhf_editor import RHFEditor


from functools import lru_cache
from typing import Any, Callable, Optional, Union, Dict, List
import uuid
from PyQt5.QtWidgets import QTabWidget, QTabBar, QWidget, QHBoxLayout, QLabel, QPushButton, QSizePolicy
from PyQt5.QtCore import Qt, QTimer
from PyQt5.QtGui import QIcon, QCursor
from ...material.system.color_manipulator import alpha
from ...base.widget_base import PyWidgetBase
from qtmui.hooks import State
from qtmui.hooks.use_theme import useTheme
from qtmui.i18n.use_translation import translate, i18n
from ...utils.validate_params import _validate_param
from qtmui.material.styles.create_theme.components.get_qss_styles import get_qss_style
from ...components.typography import Typography
from ...components.py_iconify import PyIconify
from ...components.box import Box
from .tab import Tab

class Tabs(QTabWidget, PyWidgetBase):
    """
    A tabs component, styled like Material-UI Tabs.

    The `Tabs` component manages a group of tabs with scrollable or centered layouts, aligning with
    MUI Tabs props. Inherits from native component props.

    Parameters
    ----------
    justifyContent : State, str, or None, optional
        Alignment of tabs ('flex-start', 'flex-end', qtmui-specific). Default is 'flex-start'.
        Can be a `State` object for dynamic updates.
    orientation : State, str, or None, optional
        Orientation of the tab bar ('horizontal', 'vertical'). Default is 'horizontal'.
        Can be a `State` object for dynamic updates.
    value : State, Any, or None, optional
        Value of the selected tab. Default is None.
        Can be a `State` object for dynamic updates.
    onChange : State, Callable, or None, optional
        Callback fired when the selected tab changes. Default is None.
        Can be a `State` object for dynamic updates.
    fullWidth : State or bool, optional
        If True, tabs use all available space (qtmui-specific). Default is True.
        Can be a `State` object for dynamic updates.
    fixedHeight : State, int, or None, optional
        Fixed height of the tab widget (qtmui-specific). Default is None.
        Can be a `State` object for dynamic updates.
    minWidth : State or bool, optional
        If True, sets minimum width (qtmui-specific). Default is True.
        Can be a `State` object for dynamic updates.
    children : State, List[QWidget], QWidget, or None, optional
        Content of the tabs (typically Tab components). Default is None.
        Can be a `State` object for dynamic updates.
    sx : State, List, Dict, Callable, or None, optional
        System prop for CSS overrides. Default is None.
        Can be a `State` object for dynamic updates.
    action : State, Any, or None, optional
        Callback for programmatic actions (updateIndicator, updateScrollButtons). Default is None.
        Can be a `State` object for dynamic updates.
    allowScrollButtonsMobile : State or bool, optional
        If True, scroll buttons are shown on mobile. Default is False.
        Can be a `State` object for dynamic updates.
    aria_label : State, str, or None, optional
        ARIA label for the Tabs. Default is None.
        Can be a `State` object for dynamic updates.
    aria_labelledby : State, str, or None, optional
        ARIA labelledby for the Tabs. Default is None.
        Can be a `State` object for dynamic updates.
    centered : State or bool, optional
        If True, tabs are centered. Default is False.
        Can be a `State` object for dynamic updates.
    classes : State or Dict, optional
        Override or extend styles. Default is None.
        Can be a `State` object for dynamic updates.
    component : State, str, or None, optional
        Component for the root node. Default is None.
        Can be a `State` object for dynamic updates.
    indicatorColor : State, str, or None, optional
        Color of the tab indicator ('primary', 'secondary', or custom). Default is 'primary'.
        Can be a `State` object for dynamic updates.
    scrollButtons : State, str, or bool, optional
        Behavior of scroll buttons ('auto', True, False). Default is 'auto'.
        Can be a `State` object for dynamic updates.
    selectionFollowsFocus : State or bool, optional
        If True, tab selection follows focus. Default is False.
        Can be a `State` object for dynamic updates.
    slotProps : State or Dict, optional
        Props for each slot (indicator, scrollButtons, etc.). Default is None.
        Can be a `State` object for dynamic updates.
    slots : State or Dict, optional
        Components for each slot (indicator, scrollButtons, etc.). Default is None.
        Can be a `State` object for dynamic updates.
    TabIndicatorProps : State or Dict, optional
        Props for tab indicator (deprecated, use slotProps.indicator). Default is None.
        Can be a `State` object for dynamic updates.
    TabScrollButtonProps : State or Dict, optional
        Props for scroll buttons (deprecated, use slotProps.scrollButtons). Default is None.
        Can be a `State` object for dynamic updates.
    textColor : State, str, or None, optional
        Color of tab text ('inherit', 'primary', 'secondary'). Default is 'primary'.
        Can be a `State` object for dynamic updates.
    variant : State, str, or None, optional
        Display behavior ('fullWidth', 'scrollable', 'standard'). Default is 'standard'.
        Can be a `State` object for dynamic updates.
    visibleScrollbar : State or bool, optional
        If True, scrollbar is visible. Default is False.
        Can be a `State` object for dynamic updates.
    **kwargs
        Additional keyword arguments passed to QTabWidget, supporting native props.

    Signals
    -------
    currentChanged : Signal(int)
        Emitted when the selected tab changes.

    Notes
    -----
    - `justifyContent`, `fixedHeight`, `minWidth` are qtmui-specific features.
    - `ScrollButtonComponent`, `TabIndicatorProps`, `TabScrollButtonProps` are deprecated; use `slotProps` and `slots`.
    - Supports dynamic updates via State objects.
    - MUI classes applied: `MuiTabs-root`.

    Demos:
    - Tabs: https://qtmui.com/material-ui/qtmui-tabs/

    API Reference:
    - Tabs API: https://qtmui.com/material-ui/api/tabs/
    """

    VALID_ORIENTATIONS = ['horizontal', 'vertical']
    VALID_COLORS = ['primary', 'secondary', 'inherit']
    VALID_VARIANTS = ['fullWidth', 'scrollable', 'standard']
    VALID_SCROLL_BUTTONS = ['auto', True, False]

    def __init__(
        self,
        justifyContent: Optional[Union[State, str, None]]="flex-start",
        orientation: Optional[Union[State, str, None]]="horizontal",
        value: Optional[Union[State, Any, None]]=None,
        onChange: Optional[Union[State, Callable, None]]=None,
        fullWidth: Union[State, bool]=True,
        fixedHeight: Optional[Union[State, int, None]]=None,
        minWidth: Union[State, bool]=True,
        children: Optional[Union[State, List[QWidget], QWidget, None]]=None,
        sx: Optional[Union[State, List, Dict, Callable, None]]=None,
        action: Optional[Union[State, Any, None]]=None,
        allowScrollButtonsMobile: Union[State, bool]=False,
        aria_label: Optional[Union[State, str, None]]=None,
        aria_labelledby: Optional[Union[State, str, None]]=None,
        centered: Union[State, bool]=False,
        classes: Optional[Union[State, Dict, None]]=None,
        component: Optional[Union[State, str, None]]=None,
        indicatorColor: Optional[Union[State, str, None]]="primary",
        scrollButtons: Union[State, str, bool]="auto",
        selectionFollowsFocus: Union[State, bool]=False,
        slotProps: Optional[Union[State, Dict, None]]=None,
        slots: Optional[Union[State, Dict, None]]=None,
        TabIndicatorProps: Optional[Union[State, Dict, None]]=None,
        TabScrollButtonProps: Optional[Union[State, Dict, None]]=None,
        textColor: Optional[Union[State, str, None]]="primary",
        variant: Optional[Union[State, str, None]]="standard",
        visibleScrollbar: Union[State, bool]=False,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.setObjectName(str(uuid.uuid4()))
        PyWidgetBase._setUpUi(self)
        self.theme = useTheme()
        self._widget_references = []

        # Set properties with validation
        self._set_justifyContent(justifyContent)
        self._set_orientation(orientation)
        self._set_value(value)
        self._set_onChange(onChange)
        self._set_fullWidth(fullWidth)
        self._set_fixedHeight(fixedHeight)
        self._set_minWidth(minWidth)
        self._set_children(children)
        self._set_sx(sx)
        self._set_action(action)
        self._set_allowScrollButtonsMobile(allowScrollButtonsMobile)
        self._set_aria_label(aria_label)
        self._set_aria_labelledby(aria_labelledby)
        self._set_centered(centered)
        self._set_classes(classes)
        self._set_component(component)
        self._set_indicatorColor(indicatorColor)
        self._set_scrollButtons(scrollButtons)
        self._set_selectionFollowsFocus(selectionFollowsFocus)
        self._set_slotProps(slotProps)
        self._set_slots(slots)
        self._set_TabIndicatorProps(TabIndicatorProps)
        self._set_TabScrollButtonProps(TabScrollButtonProps)
        self._set_textColor(textColor)
        self._set_variant(variant)
        self._set_visibleScrollbar(visibleScrollbar)

        self.keys = []
        self._init_ui()
        self._set_stylesheet()
        self.useEffect(
            self._set_stylesheet,
            [theme.state]
        )
        self.destroyed.connect(self._on_destroyed)
        i18n.langChanged.connect(self.retranslateUi)

    # Setter and Getter methods
    @_validate_param(file_path="qtmui.material.tabs", param_name="justifyContent", supported_signatures=Union[State, str, type(None)])
    def _set_justifyContent(self, value):
        self._justifyContent = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_justifyContent(self):
        return self._justifyContent.value if isinstance(self._justifyContent, State) else self._justifyContent

    @_validate_param(file_path="qtmui.material.tabs", param_name="orientation", supported_signatures=Union[State, str, type(None)], valid_values=VALID_ORIENTATIONS)
    def _set_orientation(self, value):
        self._orientation = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_orientation(self):
        orientation = self._orientation.value if isinstance(self._orientation, State) else self._orientation
        return orientation if orientation in self.VALID_ORIENTATIONS else 'horizontal'

    @_validate_param(file_path="qtmui.material.tabs", param_name="value", supported_signatures=Union[State, Any, type(None)])
    def _set_value(self, value):
        self._value = value
        if isinstance(value, State):
            value.valueChanged.connect(self.set_current_tab_by_key)

    def _get_value(self):
        return self._value.value if isinstance(self._value, State) else self._value

    @_validate_param(file_path="qtmui.material.tabs", param_name="onChange", supported_signatures=Union[State, Callable, type(None)])
    def _set_onChange(self, value):
        self._onChange = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_onChange(self):
        return self._onChange.value if isinstance(self._onChange, State) else self._onChange

    @_validate_param(file_path="qtmui.material.tabs", param_name="fullWidth", supported_signatures=Union[State, bool])
    def _set_fullWidth(self, value):
        self._fullWidth = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_fullWidth(self):
        return self._fullWidth.value if isinstance(self._fullWidth, State) else self._fullWidth

    @_validate_param(file_path="qtmui.material.tabs", param_name="fixedHeight", supported_signatures=Union[State, int, type(None)])
    def _set_fixedHeight(self, value):
        self._fixedHeight = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_fixedHeight(self):
        return self._fixedHeight.value if isinstance(self._fixedHeight, State) else self._fixedHeight

    @_validate_param(file_path="qtmui.material.tabs", param_name="minWidth", supported_signatures=Union[State, bool])
    def _set_minWidth(self, value):
        self._minWidth = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_minWidth(self):
        return self._minWidth.value if isinstance(self._minWidth, State) else self._minWidth

    @_validate_param(file_path="qtmui.material.tabs", param_name="children", supported_signatures=Union[State, List[QWidget], QWidget, type(None)])
    def _set_children(self, value):
        self._children = value
        if isinstance(value, QWidget):
            self._widget_references.append(value)
        elif isinstance(value, list):
            self._widget_references.extend(value)
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_children(self):
        return self._children.value if isinstance(self._children, State) else self._children

    @_validate_param(file_path="qtmui.material.tabs", param_name="sx", supported_signatures=Union[State, List, Dict, Callable, type(None)])
    def _set_sx(self, value):
        self._sx = value
        if isinstance(value, State):
            value.valueChanged.connect(self._set_stylesheet)

    def _get_sx(self):
        return self._sx.value if isinstance(self._sx, State) else self._sx

    @_validate_param(file_path="qtmui.material.tabs", param_name="action", supported_signatures=Union[State, Any, type(None)])
    def _set_action(self, value):
        self._action = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_action(self):
        return self._action.value if isinstance(self._action, State) else self._action

    @_validate_param(file_path="qtmui.material.tabs", param_name="allowScrollButtonsMobile", supported_signatures=Union[State, bool])
    def _set_allowScrollButtonsMobile(self, value):
        self._allowScrollButtonsMobile = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_allowScrollButtonsMobile(self):
        return self._allowScrollButtonsMobile.value if isinstance(self._allowScrollButtonsMobile, State) else self._allowScrollButtonsMobile

    @_validate_param(file_path="qtmui.material.tabs", param_name="aria_label", supported_signatures=Union[State, str, type(None)])
    def _set_aria_label(self, value):
        self._aria_label = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_aria_label(self):
        return self._aria_label.value if isinstance(self._aria_label, State) else self._aria_label

    @_validate_param(file_path="qtmui.material.tabs", param_name="aria_labelledby", supported_signatures=Union[State, str, type(None)])
    def _set_aria_labelledby(self, value):
        self._aria_labelledby = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_aria_labelledby(self):
        return self._aria_labelledby.value if isinstance(self._aria_labelledby, State) else self._aria_labelledby

    @_validate_param(file_path="qtmui.material.tabs", param_name="centered", supported_signatures=Union[State, bool])
    def _set_centered(self, value):
        self._centered = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_centered(self):
        return self._centered.value if isinstance(self._centered, State) else self._centered

    @_validate_param(file_path="qtmui.material.tabs", param_name="classes", supported_signatures=Union[State, Dict, type(None)])
    def _set_classes(self, value):
        self._classes = value
        if isinstance(value, State):
            value.valueChanged.connect(self._set_stylesheet)

    def _get_classes(self):
        return self._classes.value if isinstance(self._classes, State) else self._classes

    @_validate_param(file_path="qtmui.material.tabs", param_name="component", supported_signatures=Union[State, str, type(None)])
    def _set_component(self, value):
        self._component = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_component(self):
        return self._component.value if isinstance(self._component, State) else self._component

    @_validate_param(file_path="qtmui.material.tabs", param_name="indicatorColor", supported_signatures=Union[State, str, type(None)], valid_values=VALID_COLORS)
    def _set_indicatorColor(self, value):
        self._indicatorColor = value
        if isinstance(value, State):
            value.valueChanged.connect(self._set_stylesheet)

    def _get_indicatorColor(self):
        color = self._indicatorColor.value if isinstance(self._indicatorColor, State) else self._indicatorColor
        return color if color in self.VALID_COLORS or isinstance(color, str) else 'primary'

    @_validate_param(file_path="qtmui.material.tabs", param_name="scrollButtons", supported_signatures=Union[State, str, bool], valid_values=VALID_SCROLL_BUTTONS)
    def _set_scrollButtons(self, value):
        self._scrollButtons = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_scrollButtons(self):
        scroll = self._scrollButtons.value if isinstance(self._scrollButtons, State) else self._scrollButtons
        return scroll if scroll in self.VALID_SCROLL_BUTTONS else 'auto'

    @_validate_param(file_path="qtmui.material.tabs", param_name="selectionFollowsFocus", supported_signatures=Union[State, bool])
    def _set_selectionFollowsFocus(self, value):
        self._selectionFollowsFocus = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_selectionFollowsFocus(self):
        return self._selectionFollowsFocus.value if isinstance(self._selectionFollowsFocus, State) else self._selectionFollowsFocus

    @_validate_param(file_path="qtmui.material.tabs", param_name="slotProps", supported_signatures=Union[State, Dict, type(None)])
    def _set_slotProps(self, value):
        self._slotProps = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_slotProps(self):
        return self._slotProps.value if isinstance(self._slotProps, State) else self._slotProps

    @_validate_param(file_path="qtmui.material.tabs", param_name="slots", supported_signatures=Union[State, Dict, type(None)])
    def _set_slots(self, value):
        self._slots = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_slots(self):
        return self._slots.value if isinstance(self._slots, State) else self._slots

    @_validate_param(file_path="qtmui.material.tabs", param_name="TabIndicatorProps", supported_signatures=Union[State, Dict, type(None)])
    def _set_TabIndicatorProps(self, value):
        self._TabIndicatorProps = value
        if isinstance(value, State):
            value.valueChanged.connect(self._set_stylesheet)

    def _get_TabIndicatorProps(self):
        return self._TabIndicatorProps.value if isinstance(self._TabIndicatorProps, State) else self._TabIndicatorProps

    @_validate_param(file_path="qtmui.material.tabs", param_name="TabScrollButtonProps", supported_signatures=Union[State, Dict, type(None)])
    def _set_TabScrollButtonProps(self, value):
        self._TabScrollButtonProps = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_TabScrollButtonProps(self):
        return self._TabScrollButtonProps.value if isinstance(self._TabScrollButtonProps, State) else self._TabScrollButtonProps

    @_validate_param(file_path="qtmui.material.tabs", param_name="textColor", supported_signatures=Union[State, str, type(None)], valid_values=VALID_COLORS)
    def _set_textColor(self, value):
        self._textColor = value
        if isinstance(value, State):
            value.valueChanged.connect(self._set_stylesheet)

    def _get_textColor(self):
        color = self._textColor.value if isinstance(self._textColor, State) else self._textColor
        return color if color in self.VALID_COLORS else 'primary'

    @_validate_param(file_path="qtmui.material.tabs", param_name="variant", supported_signatures=Union[State, str, type(None)], valid_values=VALID_VARIANTS)
    def _set_variant(self, value):
        self._variant = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_variant(self):
        variant = self._variant.value if isinstance(self._variant, State) else self._variant
        return variant if variant in self.VALID_VARIANTS else 'standard'

    @_validate_param(file_path="qtmui.material.tabs", param_name="visibleScrollbar", supported_signatures=Union[State, bool])
    def _set_visibleScrollbar(self, value):
        self._visibleScrollbar = value
        if isinstance(value, State):
            value.valueChanged.connect(self.update_ui)

    def _get_visibleScrollbar(self):
        return self._visibleScrollbar.value if isinstance(self._visibleScrollbar, State) else self._visibleScrollbar

    def _init_ui(self):
        # Clear existing tabs
        while self.count():
            widget = self.widget(0)
            self.removeTab(0)
            if widget:
                widget.deleteLater()

        # Apply layout settings
        if self._get_fixedHeight():
            self.setFixedHeight(self._get_fixedHeight())
        if self._get_minWidth():
            self.setMinimumWidth(200)  # Arbitrary default minWidth
        if self._get_orientation() == "vertical":
            self.setTabPosition(QTabWidget.West)
        else:
            self.setTabPosition(QTabWidget.North)

        # Apply ARIA attributes
        if self._get_aria_label():
            self.setProperty("aria-label", self._get_aria_label())
        if self._get_aria_labelledby():
            self.setProperty("aria-labelledby", self._get_aria_labelledby())

        # Apply centered or justifyContent
        if self._get_centered():
            self.tabBar().setProperty("centered", "true")
        elif self._get_justifyContent() == "flex-end":
            self.setLayoutDirection(Qt.RightToLeft)
        else:
            self.setLayoutDirection(Qt.LeftToRight)

        # Apply variant
        if self._get_variant() == "fullWidth" or self._get_fullWidth():
            self.tabBar().setExpanding(True)
        else:
            self.tabBar().setExpanding(False)

        # Initialize scroll buttons
        scroll_buttons = self._get_scrollButtons()
        show_scroll = scroll_buttons == True or (scroll_buttons == "auto" and self._get_allowScrollButtonsMobile())
        self.setUsesScrollButtons(show_scroll)
        self.setDocumentMode(self._get_variant() == "scrollable")

        # Initialize children
        children = self._get_children()
        if isinstance(children, list):
            for index, tab in enumerate(children):
                if isinstance(tab, Tab):
                    self.addTab(tab, "")
                    self.set_tab_with_icon_at_position(index, tab._get_label(), tab._get_icon(), tab._get_iconPosition())
                    if tab._get_key():
                        self.keys.append(tab._get_key())

        # Set cursor
        self.tabBar().setCursor(QCursor(Qt.PointingHandCursor))

        # Connect signals
        self.currentChanged.connect(self.on_tab_changed)
        if self._get_onChange():
            self.currentChanged.connect(lambda index: self._get_onChange()(self.currentWidget()._get_key()))

        # Apply selectionFollowsFocus
        if self._get_selectionFollowsFocus():
            self.tabBar().focusInEvent = lambda event: self.setCurrentIndex(self.tabBar().currentIndex())

        # Set current tab
        if self._get_value():
            self.set_current_tab_by_key(self._get_value())

        # Apply action
        if self._get_action():
            QTimer.singleShot(0, lambda: self._get_action()(self))

    def set_tab_with_icon_at_position(self, index, title, icon, iconPosition="top"):
        custom_widget = QWidget(self)
        layout = QVBoxLayout(custom_widget) if iconPosition in ["top", "bottom"] else QHBoxLayout(custom_widget)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(self.theme.spacing(1))

        # Create label
        label_text = Typography(variant="body1", text=title, color=self._get_textColor()) if title else None
        if label_text:
            label_text.setParent(custom_widget)
            self._widget_references.append(label_text)

        # Create icon
        label_icon = None
        if isinstance(icon, str):
            label_icon = PyIconify(key=icon, size=20, color=self.theme.palette[self._get_textColor()].main)
            label_icon.setParent(custom_widget)
            self._widget_references.append(label_icon)
        elif isinstance(icon, QWidget):
            label_icon = icon
            label_icon.setParent(custom_widget)

        # Arrange icon and text
        if iconPosition == "top":
            if label_icon:
                layout.addWidget(label_icon)
            if label_text:
                layout.addWidget(label_text)
        elif iconPosition == "bottom":
            if label_text:
                layout.addWidget(label_text)
            if label_icon:
                layout.addWidget(label_icon)
        elif iconPosition == "start":
            if label_icon:
                layout.addWidget(label_icon)
            if label_text:
                layout.addWidget(label_text)
        elif iconPosition == "end":
            if label_text:
                layout.addWidget(label_text)
            if label_icon:
                layout.addWidget(label_icon)

        layout.setAlignment(Qt.AlignCenter)
        custom_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.tabBar().setTabButton(index, QTabBar.RightSide, custom_widget)

    @classmethod
    @lru_cache(maxsize=128)
    def _get_stylesheet(cls, _theme_mode: str) -> str:
        theme = useTheme()
        tabs_styles = theme.components.get("Tabs", {}).get("styles", {})
        tab_bar_styles = theme.components.get("TabBar", {}).get("styles", {})
        
        tabs_root_qss = get_qss_style(tabs_styles.get("root", {}))
        tabs_pane_qss = get_qss_style(tabs_styles.get("pane", {}))
        tab_bar_root_qss = get_qss_style(tab_bar_styles.get("root", {}))
        tab_bar_tab_qss = get_qss_style(tab_bar_styles.get("tab", {}))
        tab_bar_tab_selected_qss = get_qss_style(tab_bar_styles.get("tab", {}).get("slots", {}).get("selected", {}))
        tab_bar_tab_not_selected_qss = get_qss_style(tab_bar_styles.get("tab", {}).get("slots", {}).get("notSelected", {}))

        stylesheet = f"""
            QTabWidget {{
                {tabs_root_qss}
            }}
            QTabWidget::pane {{
                {tabs_pane_qss}
            }}
            QTabBar {{
                {tab_bar_root_qss}
            }}
            QTabBar::tab {{
                {tab_bar_tab_qss}
                color: {theme.palette[self._get_textColor()].main};
            }}
            QTabBar::tab:selected {{
                {tab_bar_tab_selected_qss}
                border-bottom: 2px solid {theme.palette[self._get_indicatorColor()].main};
            }}
            QTabBar::tab:!selected {{
                {tab_bar_tab_not_selected_qss}
            }}
            QTabBar[centered=true] {{
                alignment: center;
            }}
        """
        return stylesheet

    def _set_stylesheet(self):
        _theme_mode = self.theme.palette.mode
        stylesheet = self._get_stylesheet(_theme_mode)
        
        # Handle classes
        classes = self._get_classes() or {}
        classes_qss = get_qss_style(classes, class_name=f"#{self.objectName()}")

        # Handle sx
        sx = self._get_sx()
        sx_qss = ""
        if sx:
            if isinstance(sx, (list, dict)):
                sx_qss = get_qss_style(sx, class_name=f"#{self.objectName()}")
            elif isinstance(sx, Callable):
                sx_result = sx()
                if isinstance(sx_result, (list, dict)):
                    sx_qss = get_qss_style(sx_result, class_name=f"#{self.objectName()}")
                elif isinstance(sx_result, str):
                    sx_qss = sx_result
            elif isinstance(sx, str) and sx != "":
                sx_qss = sx

        # Apply variant and scrollbar
        variant_qss = ""
        if self._get_variant() == "scrollable" or self._get_visibleScrollbar():
            variant_qss = "QTabBar { overflow-x: auto; }"

        self.setStyleSheet(f"""
            {stylesheet}
            #{self.objectName()} {{
                {classes_qss}
                {variant_qss}
            }}
            {sx_qss}
        """)

    def set_current_tab_by_key(self, key):
        for i in range(self.count()):
            tab = self.widget(i)
            if getattr(tab, '_get_key', lambda: None)() == key:
                self.setCurrentIndex(i)
                return True
        return False

    def retranslateUi(self):
        for i in range(self.count()):
            tab = self.widget(i)
            if isinstance(tab, Tab):
                current_text = tab._get_label() or f"Tab {i + 1}"
                self.setTabText(i, translate(current_text))

    def on_tab_changed(self, index):
        current_tab = self.widget(index)
        if current_tab:
            current_tab._init_ui()

    def update_ui(self):
        self._init_ui()
        self._set_stylesheet()

    def _on_destroyed(self):
        self._widget_references.clear()

    def disable_pointer_events(self):
        self.setCursor(Qt.BlankCursor)

    def paintEvent(self, event):
        PyWidgetBase.paintEvent(self, event)
        return super().paintEvent(event)

    def resizeEvent(self, event):
        PyWidgetBase.resizeEvent(self, event)
        return super().resizeEvent(event)


# qtmui/material/stack.py
import asyncio
from functools import lru_cache

import threading
from typing import Callable, Optional, Union, Dict, List, Any
from PySide6.QtWidgets import QFrame, QSpacerItem, QSizePolicy, QWidget, QGraphicsDropShadowEffect, QBoxLayout, QApplication
from PySide6.QtCore import Qt, QTimer, Signal, QRunnable, QThreadPool
from PySide6.QtGui import QColor, QPainter
import uuid
import re
from qtmui.hooks import State
from qtmui.material.styles import useTheme
from qtmui.material.styles.create_theme.components.get_qss_styles import get_qss_style
from qtmui.material.widget_base import PyWidgetBase
from qtmui.utils.data import deep_merge
from qtmui.configs import LOAD_WIDGET_ASYNC

# from ..flow_layout import FlowLayout
from ..layouts.flow_layout import FlowLayout

from qtmui.material.utils.sx_helper import padding_to_qmargins_args, get_sx_dict, get_padding_tuple

from qtmui.material.utils.style_decorator import StyleDecorator

from ..utils.validate_params import _validate_param

class Stack(QFrame, PyWidgetBase):
    """
    A stack component, styled like Material-UI Stack.

    The `Stack` component manages layout of immediate children in a flexbox-like manner,
    supporting direction, spacing, alignment, and dividers, aligning with MUI Stack props.
    Inherits from native component props.

    Parameters
    ----------
    parent : QWidget, optional
        The parent widget. Default is None.
    direction : State, str, List[str], or Dict, optional
        Defines the flex-direction ('row', 'row-reverse', 'column', 'column-reverse').
        Supports arrays or objects for responsive design. Default is 'column'.
        Can be a `State` object for dynamic updates.
    spacing : State, int, float, List, or Dict, optional
        Defines the space between immediate children. Default is 0.
        Can be a `State` object for dynamic updates.
    divider : State, QWidget, or None, optional
        Adds an element between each child. Default is None.
        Can be a `State` object for dynamic updates.
    hightLight : State or bool, optional
        If True, applies a highlight effect (qtmui-specific). Default is False.
        Can be a `State` object for dynamic updates.
    key : State or Any, optional
        Unique key identifier. Default is None.
        Can be a `State` object for dynamic updates.
    alignItems : State or str, optional
        Defines the align-items style ('flex-start', 'center', 'flex-end', 'stretch', 'baseline').
        Default is 'stretch'.
        Can be a `State` object for dynamic updates.
    justifyContent : State or str, optional
        Defines the justify-content style ('flex-start', 'flex-end', 'center', 'space-between',
        'space-around', 'space-evenly'). Default is 'flex-start'.
        Can be a `State` object for dynamic updates.
    flexWrap : State or str, optional
        Defines the flex-wrap style ('nowrap', 'wrap'). Default is 'nowrap'.
        Can be a `State` object for dynamic updates.
    flexGrow : State, int, or None, optional
        Defines the flex-grow factor. Default is None.
        Can be a `State` object for dynamic updates.
    children : State, List[QWidget], QWidget, or None, optional
        List of child elements. Default is None.
        Can be a `State` object for dynamic updates.
    variant : State, str, or None, optional
        Variant style (qtmui-specific, e.g., 'outlined'). Default is None.
        Can be a `State` object for dynamic updates.
    sx : State, List, Dict, Callable, or None, optional
        System prop for CSS overrides. Default is None.
        Can be a `State` object for dynamic updates.
    component : State, str, type, or None, optional
        Component used for the root node. Default is None (uses QFrame).
        Can be a `State` object for dynamic updates.
    useFlexGap : State or bool, optional
        If True, uses CSS flexbox gap instead of margin. Default is False.
        Can be a `State` object for dynamic updates.
    **kwargs
        Additional keyword arguments passed to QFrame, supporting native component props.

    Notes
    -----
    - Existing parameters (13) are retained; 2 new parameters added to align with MUI Stack.
    - `hightLight` and `variant` are qtmui-specific and not part of MUI Stack.
    - Supports responsive direction and spacing via arrays or objects.
    - MUI classes applied: `MuiStack-root`.

    Demos:
    - Stack: https://qtmui.com/material-ui/qtmui-stack/

    API Reference:
    - Stack API: https://qtmui.com/material-ui/api/stack/
    """

    updateStyleSheet = Signal(object)


    VALID_DIRECTIONS = ["row", "row-reverse", "column", "column-reverse"]  # Chuyển thành list để khớp với _validate_param
    VALID_ALIGN_ITEMS = ["flex-start", "center", "flex-end", "stretch", "baseline"]
    VALID_JUSTIFY_CONTENT = ["flex-start", "flex-end", "center", "space-between", "space-around", "space-evenly"]
    VALID_FLEX_WRAP = ["nowrap", "wrap"]
    VALID_VARIANTS = [None, "outlined"]

    def __init__(
        self,
        parent: Optional[QWidget] = None,
        direction: Optional[Union[State, str, dict]] = "column",  # Direction for arranging child elements
        spacing: Optional[Union[State, int, float, dict]] = 0,  # Spacing between child elements
        divider: Optional[Union[State, Callable, QWidget, bool]] = None,  # Divider between child elements
        hightLight: Optional[Union[State, bool]] = False,  # Highlight effect
        key: Optional[Union[State, str]] = None,  # Key identifier
        alignItems: Optional[Union[State, str]] = "stretch",  # Alignment of child elements
        justifyContent: Optional[Union[State, str]] = "flex-start",  # Justification of child elements
        flexWrap: Optional[Union[State, str]] = "nowrap",  # Flex wrap behavior
        flexGrow: Optional[Union[State, int]] = None,  # Flex grow factor
        children: Optional[Union[State, List[QWidget]]] = None,  # List of child elements
        variant: Optional[Union[State, str]] = None,  # Variant style
        sx: Optional[Union[State, Callable, str, Dict]] = None,  # Custom styles
        asynRenderQss: Optional[Union[State, bool]] = False,
        **kwargs
    ):
        super().__init__(parent)
        self.setObjectName(str(uuid.uuid4()))  # Gán objectName trước khi gọi set_sx
        
        self._flowLayoutPadding = None
        self._repaint_scheduled = False
        
        # self._sx: Dict[str, Any] = sx or {"border": "1px solid rgba(0,0,0,100)"}
        # self._decor = StyleDecorator(self, self._sx)
        
        
        if sx:
            # Nếu flexWrap thì dùng FlowLayout, lúc này border và padding sẽ bị bóc ra khỏi sx để paint border, 
            # padding truyền cho FlowLayout
            strFlexWrap = "nowrap"
            if isinstance(flexWrap, State):
                strFlexWrap = flexWrap.value
            elif isinstance(flexWrap, str):
                strFlexWrap = flexWrap
            
            if strFlexWrap == "wrap":
                sx_dict = get_sx_dict(sx)
                self._flowLayoutPadding = get_padding_tuple(sx_dict)
                self._decor = StyleDecorator(self, sx_dict)
                
                sx.update({"padding": "0px"})
                sx.update({"border-width": "0px"})
            
            self._setSx(sx)
        self._setKwargs(kwargs)
        
        self._kwargs = kwargs.copy()
        
        
        PyWidgetBase._setUpUi(self)

        self.theme = useTheme()

        # List of child components with position: absolute
        self._child_abs = []
        self.breakpoint = None
        

        # Gán giá trị ban đầu cho các thuộc tính bằng các hàm _set_*
        self._set_direction(direction)
        self._set_spacing(spacing)
        self._set_divider(divider)
        self._set_hightLight(hightLight)
        self._set_key(key)
        self._set_alignItems(alignItems)
        self._set_justifyContent(justifyContent)
        self._set_flexWrap(flexWrap)
        self._set_flexGrow(flexGrow)
        self._set_children(children or [])
        self._set_variant(variant)
        self._set_sx(sx)

        # from PyWidgetBase
        self._setup_sx_position(sx)  # Gán sx và khởi tạo các thuộc tính định vị

        self._asynRenderQss = asynRenderQss

        self._init_ui()


    # @_validate_param(file_path="qtmui.material.stack", param_name="direction", supported_signatures=Union[State, dict, str], valid_values=VALID_DIRECTIONS)
    def _set_direction(self, value):
        """Assign value to direction."""
        self._direction = value

    def _get_direction(self):
        """Get the direction value."""
        return self._direction.value if isinstance(self._direction, State) else self._direction

    # @_validate_param(file_path="qtmui.material.stack", param_name="spacing", supported_signatures=Union[State, int, float], validator=lambda x: x >= 0)
    def _set_spacing(self, value):
        """Assign value to spacing."""
        self._spacing = value

    def _get_spacing(self):
        """Get the spacing value."""
        spacing_value = self._spacing.value if isinstance(self._spacing, State) else self._spacing
        if (isinstance(spacing_value, int) or isinstance(spacing_value, float)):
            spacing_value = spacing_value * self.theme.spacing.default_spacing
        
        return spacing_value
    
    def _get_list_padding(self):
        """Get the spacing value."""
        padding_value = self._get_sx_value("padding")
        padding_left_value = self._get_sx_value("padding-left")
        if isinstance(padding_value, str) and padding_value.endswith("px"):
            padding_value = int(padding_value.replace("px", "").strip())
        spacing_value = self._spacing.value if isinstance(self._spacing, State) else self._spacing
        if (isinstance(spacing_value, int) or isinstance(spacing_value, float)):
            spacing_value = spacing_value * self.theme.spacing.default_spacing
        
        return spacing_value

    @_validate_param(file_path="qtmui.material.stack", param_name="divider", supported_signatures=Union[State, Callable, QWidget, bool, type(None)])
    def _set_divider(self, value):
        """Assign value to divider."""
        self._divider = value

    def _get_divider(self):
        """Get the divider value."""
        return self._divider.value if isinstance(self._divider, State) else self._divider

    @_validate_param(file_path="qtmui.material.stack", param_name="hightLight", supported_signatures=Union[State, bool])
    def _set_hightLight(self, value):
        """Assign value to hightLight."""
        self._hightLight = value

    def _get_hightLight(self):
        """Get the hightLight value."""
        return self._hightLight.value if isinstance(self._hightLight, State) else self._hightLight

    # @_validate_param(file_path="qtmui.material.stack", param_name="key", supported_signatures=Union[State, str, type(None)])
    def _set_key(self, value):
        """Assign value to key."""
        self._key = value

    def _get_key(self):
        """Get the key value."""
        return self._key.value if isinstance(self._key, State) else self._key

    @_validate_param(file_path="qtmui.material.stack", param_name="alignItems", supported_signatures=Union[State, str], valid_values=VALID_ALIGN_ITEMS)
    def _set_alignItems(self, value):
        """Assign value to alignItems."""
        self._alignItems = value

    def _get_alignItems(self):
        """Get the alignItems value."""
        return self._alignItems.value if isinstance(self._alignItems, State) else self._alignItems

    @_validate_param(file_path="qtmui.material.stack", param_name="justifyContent", supported_signatures=Union[State, str], valid_values=VALID_JUSTIFY_CONTENT)
    def _set_justifyContent(self, value):
        """Assign value to justifyContent."""
        self._justifyContent = value

    def _get_justifyContent(self):
        """Get the justifyContent value."""
        return self._justifyContent.value if isinstance(self._justifyContent, State) else self._justifyContent

    @_validate_param(file_path="qtmui.material.stack", param_name="flexWrap", supported_signatures=Union[State, str], valid_values=VALID_FLEX_WRAP)
    def _set_flexWrap(self, value):
        """Assign value to flexWrap."""
        self._flexWrap = value

    def _get_flexWrap(self):
        """Get the flexWrap value."""
        return self._flexWrap.value if isinstance(self._flexWrap, State) else self._flexWrap

    @_validate_param(file_path="qtmui.material.stack", param_name="flexGrow", supported_signatures=Union[State, int, type(None)], validator=lambda x: x >= 0 if x is not None else True)
    def _set_flexGrow(self, value):
        """Assign value to flexGrow."""
        self._flexGrow = value

    def _get_flexGrow(self):
        """Get the flexGrow value."""
        return self._flexGrow.value if isinstance(self._flexGrow, State) else self._flexGrow

    @_validate_param(file_path="qtmui.material.stack", param_name="children", supported_signatures=Union[State, list])
    def _set_children(self, value):
        """Assign value to children."""
        self._children = value

    def _get_children(self):
        """Get the list of children."""
        return self._children.value if isinstance(self._children, State) else self._children

    @_validate_param(file_path="qtmui.material.stack", param_name="variant", supported_signatures=Union[State, str, type(None)], valid_values=VALID_VARIANTS)
    def _set_variant(self, value):
        """Assign value to variant."""
        self._variant = value

    def _get_variant(self):
        """Get the variant value."""
        return self._variant.value if isinstance(self._variant, State) else self._variant

    @_validate_param(file_path="qtmui.material.stack", param_name="sx", supported_signatures=Union[State, Callable, str, Dict, type(None)])
    def _set_sx(self, value):
        """Assign value"""
        self._sx = value

    def _init_ui(self):
        """Initialize the Stack UI."""
        self.theme = useTheme()

        # Connect signals for State parameters
        self._connect_signals()

        self._update_layout()

        if self._tooltip:
            PyWidgetBase._installTooltipFilter(self)

        # Update positions of child components with position: absolute
        PyWidgetBase.update_absolute_children(self)

        self.theme.state.valueChanged.connect(self._onThemeChanged)
        # QTimer.singleShot(0, self._scheduleSetStyleSheet)
        if self._asynRenderQss:
            self.updateStyleSheet.connect(self._updateStylesheet)
        else:
            self._setStyleSheet()
        
        self.destroyed.connect(lambda obj: self._onDestroy())

        # Connect signals for State parameters
        self._connect_signals()

    def _onDestroy(self, obj=None):
        # Cancel task nếu đang chạy
        if hasattr(self, "_setupStyleSheet") and self._setupStyleSheet and not self._setupStyleSheet.done():
            self._setupStyleSheet.cancel()

    def _onThemeChanged(self):
        if not self.isVisible():
            return
        QTimer.singleShot(0, self._scheduleSetStyleSheet)

    def _scheduleSetStyleSheet(self):
        self._setupStyleSheet = asyncio.ensure_future(self._lazy_setStyleSheet())

    async def _lazy_setStyleSheet(self):
        self._setStyleSheet()

    @classmethod
    @lru_cache(maxsize=128)
    def _getStyleSheet(cls, objectName: str, styledConfig: str="Stack"):
        theme = useTheme()
        if hasattr(cls, "styledDict"):
            themeComponent = deep_merge(theme.components, cls.styledDict)
        else:
            themeComponent = theme.components
            
        PyStack_root = themeComponent["PyStack"].get("styles")["root"](cls.ownerState)
        PyStack_root_qss = get_qss_style(PyStack_root, class_name=f"Stack")
        PyStack_root_prop_outlinedVariant_qss = get_qss_style(PyStack_root["props"]["outlinedVariant"])
        # print('PyStack_root_prop_outlinedVariant_qss_____________________', PyStack_root_prop_outlinedVariant_qss)

        stylesheet = f"""
            {PyStack_root_qss}
            #{objectName}[variant=outlined] {{
                {PyStack_root_prop_outlinedVariant_qss}
            }}
        """

        return stylesheet
    
    def _renderStylesheet(self):
        stylesheet = ""
        if hasattr(self, "styledDict"):
            root = self.styledDict.get("PyStack", {}).get("styles", {}).get("root", None)(self._kwargs)
            if root:
                stylesheet = self._getStyleSheet(objectName=self.objectName(), styledConfig=str(root))
        else:
            stylesheet = self._getStyleSheet(objectName=self.objectName())
            
        sxQss = ""
        if self._sx:
            sxQss = self._getSxQss(sxStr=str(self._sx), className=f"#{self.objectName()}")
        stylesheet = f"""
            {stylesheet}
            {sxQss}
        """
        
        self.updateStyleSheet.emit(stylesheet)
        
    @classmethod
    def _setSx(cls, sx: dict = {}):
        cls.sxDict = sx
        
    @classmethod
    def _setKwargs(cls, kwargs: dict = {}):
        cls.ownerState = kwargs

    @classmethod
    @lru_cache(maxsize=128)
    def _getSxQss(cls, sxStr: str = "", className: str = "PyWidgetBase"):
        sx_qss = get_qss_style(cls.sxDict, class_name=className)
        return sx_qss
        
    def _updateStylesheet(self, stylesheet):
        self.setStyleSheet(stylesheet)
    
    def _setStyleSheet(self):
        stylesheet = ""
        if hasattr(self, "styledDict"):
            root = self.styledDict.get("PyStack", {}).get("styles", {}).get("root", None)(self._kwargs)
            if root:
                stylesheet = self._getStyleSheet(objectName=self.objectName(), styledConfig=str(root))
        else:
            stylesheet = self._getStyleSheet(objectName=self.objectName())
            
        sxQss = ""
        if self._sx:
            sxQss = self._getSxQss(sxStr=str(self._sx), className=f"#{self.objectName()}")

        stylesheet = f"""
            {stylesheet}
            {sxQss}
        """

        self.setStyleSheet(stylesheet)


    def _connect_signals(self):
        """Connect valueChanged signals of State parameters to their slots."""
        if isinstance(self._direction, State):
            self._direction.valueChanged.connect(self._change_direction)
        if isinstance(self._spacing, State):
            self._spacing.valueChanged.connect(self._update_spacing)
        if isinstance(self._divider, State):
            self._divider.valueChanged.connect(self._setup_layout)
        if isinstance(self._hightLight, State):
            self._hightLight.valueChanged.connect(self._update_highlight)
        if isinstance(self._key, State):
            self._key.valueChanged.connect(self._set_key)
        if isinstance(self._alignItems, State):
            self._alignItems.valueChanged.connect(self._setup_layout)
        if isinstance(self._justifyContent, State):
            self._justifyContent.valueChanged.connect(self._setup_layout)
        if isinstance(self._flexWrap, State):
            self._flexWrap.valueChanged.connect(self._update_flex_wrap)
        if isinstance(self._flexGrow, State):
            self._flexGrow.valueChanged.connect(self._setup_layout)
        if isinstance(self._children, State):
            self._children.valueChanged.connect(self._update_children)
        if isinstance(self._variant, State):
            self._variant.valueChanged.connect(self._update_variant)
        if isinstance(self._sx, State):
            self._sx.valueChanged.connect(self._setStyleSheet)



    def _update_layout(self):
        """Update the layout based on current properties."""
        _flexWrap = self._get_flexWrap()
        if _flexWrap == "wrap":
            clean_list = list(filter(lambda x: x is not None, self._get_children()))
            self.setLayout(FlowLayout(parent=self, spacing=self._get_spacing(), margin=self._flowLayoutPadding, children=clean_list, alignItems="center", justifyContent=self._get_justifyContent(), sx=self._sx))
            return

        self._layout = QBoxLayout(self._get_direction_enum(self._direction))
        self.setLayout(self._layout)
        self._update_spacing()
        self._layout.setContentsMargins(0, 0, 0, 0)

        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Maximum)

        self._setup_layout()

    def _update_spacing(self):
        """Update the spacing of the layout."""
        spacing_value = self._get_spacing()
        self._layout.setSpacing(self._resolve_breakpoint_value(spacing_value, self._get_breakpoint()))
        self.update()

    def _update_highlight(self):
        """Update highlight effect."""
        hightLight_value = self._get_hightLight()
        if hightLight_value:
            shadow = QGraphicsDropShadowEffect(self)
            shadow.setBlurRadius(5)
            shadow.setColor(QColor(0, 0, 0, 50))
            shadow.setOffset(2, 2)
            self.setGraphicsEffect(shadow)
        else:
            self.setGraphicsEffect(None)
        self.update()

    def _update_flex_wrap(self):
        """Update the flex wrap behavior."""
        self._update_layout()

    def _update_children(self):
        """Update the children of the Stack."""
        self._clear_widgets()
        self._setup_layout()

    def _update_variant(self):
        """Update the variant style."""
        variant_value = self._get_variant()
        if variant_value == "outlined":
            self.setProperty("variant", "outlined")
        self._setStyleSheet()


    def _change_direction(self):
        """Handle direction change."""
        self._layout.setDirection(self._get_direction_enum(self._direction))
        self.update()
        self._setup_layout()


    def _setup_layout(self):
        """Asynchronously set up the layout."""
        try:
            self._clear_layout()
            _direction = self._get_direction()
            self._layout.setDirection(self._get_direction_enum(_direction))

            children = self._get_children()
            if children:
                for index, widget in enumerate(children, 1):
                    if widget is None:
                        continue  # Skip if widget is None
                    # Check if the widget has position: absolute
                    sx = getattr(widget, '_sx', None)
                    if sx and isinstance(sx, dict) and sx.get("position") == "absolute":
                        self._child_abs.append(widget)
                        widget._relative_to = self

                    if LOAD_WIDGET_ASYNC:
                        self._do_task_async(lambda index=index, widget=widget: self._add_widget(index, widget))
                    else:
                        self._add_widget(index, widget)

            self.update()
            self.update_absolute_children()
        except Exception as e:
            pass
            # raise  # Ném lỗi trực tiếp mà không cần logging

    def _clear_layout(self):
        """Clear all widgets from the layout."""
        while self._layout.count():
            item = self._layout.takeAt(0)
            if item.widget():
                widget = item.widget()
                widget.setParent(None)
                # If widget has position: absolute, remove from managed list
                if widget in self._child_abs:
                    self._child_abs.remove(widget)
                    widget._relative_to = None
            elif isinstance(item, QSpacerItem):
                self._layout.removeItem(item)

    def _clear_widgets(self):
        """Clear old widgets and disconnect their signals if any."""
        children = self._get_children()
        for child in children:
            if child and hasattr(child, 'destroyed'):
                try:
                    child.destroyed.disconnect()
                except TypeError:
                    pass  # Ignore if signal is already disconnected or not connected
            if child:
                child.setParent(None)
                if child in self._child_abs:
                    self._child_abs.remove(child)
                    child._relative_to = None


    def _on_destroyed(self):
        """Disconnect signals when the object is destroyed."""
        try:
            self.theme.state.valueChanged.disconnect(self._update_stylesheet)
        except TypeError:
            pass
        self._layout = None
        self._children = None




    def _add_widget(self, index, widget: QWidget):
        """Add a widget to the layout with proper alignment and spacing."""
        _justifyContent = self._get_justifyContent()
        _direction = self._get_direction()

        if index == 1:
            for i in reversed(range(self._layout.count())):
                item = self._layout.itemAt(i)
                if isinstance(item, QSpacerItem):
                    self._layout.removeItem(item)

        if _justifyContent == "space-around":
            if index == 1:
                self._layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum) if _direction in ["row", "row-reverse"] else QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))
            self._add_widget_with_alignment_and_flex_shrink(self._layout, widget)
            if index == len(self._get_children()):
                self._layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum) if _direction in ["row", "row-reverse"] else QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))
            else:
                self._layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum) if _direction in ["row", "row-reverse"] else QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))
                if self._divider:
                    divider = self._create_divider()
                    if divider:
                        self._layout.addWidget(divider)
                self._layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum) if _direction in ["row", "row-reverse"] else QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))

        elif _justifyContent == "space-between":
            self._add_widget_with_alignment_and_flex_shrink(self._layout, widget)
            if index < len(self._get_children()):
                self._layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum) if _direction in ["row", "row-reverse"] else QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))
                if self._divider:
                    divider = self._create_divider()
                    if divider:
                        self._layout.addWidget(divider)

        elif _justifyContent == "space-evenly":
            if index == 1:
                self._layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum) if _direction in ["row", "row-reverse"] else QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))
            self._add_widget_with_alignment_and_flex_shrink(self._layout, widget)
            if index == len(self._get_children()):
                self._layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum) if _direction in ["row", "row-reverse"] else QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))
            else:
                self._layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum) if _direction in ["row", "row-reverse"] else QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))
                if self._divider:
                    divider = self._create_divider()
                    if divider:
                        self._layout.addWidget(divider)

        elif _justifyContent == "flex-start":
            if _direction == "row-reverse":
                self._add_widget_with_alignment_and_flex_shrink(self._layout, widget)
                if self._divider and index < len(self._get_children()):
                    divider = self._create_divider()
                    if divider:
                        self._layout.addWidget(divider)
                if index == len(self._get_children()):
                    self._layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))
            else:
                if index == 1 and _direction in ["row", "column", "column-reverse"]:
                    pass
                self._add_widget_with_alignment_and_flex_shrink(self._layout, widget)
                if self._divider and index < len(self._get_children()):
                    divider = self._create_divider()
                    if divider:
                        self._layout.addWidget(divider)
                if index == len(self._get_children()) and _direction == "row":
                    self._layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum))
                elif index == len(self._get_children()) and _direction not in ["row", "row-reverse"]:
                    self._layout.addItem(QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))

        elif _justifyContent == "flex-end":
            if index == 1:
                self._layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum) if _direction in ["row", "row-reverse"] else QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))
            self._add_widget_with_alignment_and_flex_shrink(self._layout, widget)
            if self._divider and index < len(self._get_children()):
                divider = self._create_divider()
                if divider:
                    self._layout.addWidget(divider)

        elif _justifyContent == "center":
            if index == 1:
                self._layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum) if _direction in ["row", "row-reverse"] else QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))
            self._add_widget_with_alignment_and_flex_shrink(self._layout, widget)
            if self._divider and index < len(self._get_children()):
                divider = self._create_divider()
                if divider:
                    self._layout.addWidget(divider)
            if index == len(self._get_children()):
                self._layout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Minimum) if _direction in ["row", "row-reverse"] else QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding))

    def _add_widget_with_alignment_and_flex_shrink(self, layout: QBoxLayout, widget: QWidget):
        """Add a widget to the layout with alignment and flex shrink considerations."""
        _direction = self._get_direction()
        _alignItems = self._get_alignItems()

        size_policy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        align_type = None

        align_self = self._get_align_self(widget)
        flex_grow = self._get_flex_grow(widget)
        flex_shrink = self._get_flex_shrink(widget)

        if _direction in ["row", "row-reverse"]:
            if align_self == "stretch" or (_alignItems == "stretch" and align_self is None):
                size_policy.setVerticalPolicy(QSizePolicy.Expanding)
            elif align_self == "baseline" or (_alignItems == "baseline" and align_self is None):
                align_type = Qt.AlignTop
            else:
                align_type = self._get_align_mapped(align_self) if align_self else {
                    "flex-start": Qt.AlignTop,
                    "center": Qt.AlignVCenter,
                    "flex-end": Qt.AlignBottom
                }.get(_alignItems, Qt.AlignVCenter)
        else:  # column or column-reverse
            if align_self == "stretch" or (_alignItems == "stretch" and align_self is None):
                size_policy.setHorizontalPolicy(QSizePolicy.Expanding)
            elif align_self == "baseline" or (_alignItems == "baseline" and align_self is None):
                align_type = Qt.AlignLeft
            else:
                align_type = self._get_align_mapped(align_self) if align_self else {
                    "flex-start": Qt.AlignLeft,
                    "center": Qt.AlignHCenter,
                    "flex-end": Qt.AlignRight
                }.get(_alignItems, Qt.AlignHCenter)

        try:
            widget.setSizePolicy(size_policy)
        except Exception as e:
            pass  # Bỏ logging, chỉ bỏ qua lỗi

        stretch = int(flex_grow or 0)
        if align_type is not None:
            if stretch > 0:
                layout.addWidget(widget, stretch=stretch, alignment=align_type)
            else:
                layout.addWidget(widget, alignment=align_type)
        else:
            if stretch > 0:
                layout.addWidget(widget, stretch=stretch)
            else:
                layout.addWidget(widget)

    def _get_align_self(self, widget: QWidget):
        """Get the alignSelf property from the widget's sx."""
        if hasattr(widget, '_sx') and widget._sx and isinstance(widget._sx, dict):
            return widget._sx.get("alignSelf")
        return None

    def _get_flex_grow(self, widget: QWidget):
        """Get the flexGrow property from the widget's sx or Stack's flexGrow."""
        flexGrow = self._get_flexGrow()
        if flexGrow:
            return flexGrow
        if hasattr(widget, '_sx') and widget._sx and isinstance(widget._sx, dict):
            return widget._sx.get("flexGrow", 0)
        return 0

    def _get_flex_shrink(self, widget: QWidget):
        """Get the flexShrink property from the widget's sx."""
        if hasattr(widget, '_sx') and widget._sx and isinstance(widget._sx, dict):
            return widget._sx.get("flexShrink", 0)
        return 0

    def _get_flex_basis(self, widget: QWidget):
        """Get the flexBasis property from the widget's sx."""
        if hasattr(widget, '_sx') and widget._sx and isinstance(widget._sx, dict):
            return widget._sx.get("flexBasis")
        return None

    def _get_direction_enum(self, direction):
        """Map direction to QBoxLayout direction."""
        direction = direction.value if isinstance(direction, State) else direction
        direction_map = {
            "row": QBoxLayout.LeftToRight,
            "row-reverse": QBoxLayout.RightToLeft,
            "column": QBoxLayout.TopToBottom,
            "column-reverse": QBoxLayout.BottomToTop
        }
        return direction_map.get(self._resolve_breakpoint_value(direction, self._get_breakpoint()), QBoxLayout.TopToBottom)

    def _change_direction(self):
        """Handle direction change."""
        self._layout.setDirection(self._get_direction_enum(self._resolve_breakpoint_value(self._direction, self._get_breakpoint())))
        self.update()
        self._setup_layout()

    def _create_divider(self):
        """Create a divider widget based on the divider property."""
        _divider = self._get_divider()
        if isinstance(_divider, Callable):
            divider = _divider()
            if isinstance(divider, QWidget):
                return divider
            raise TypeError(f"The return result of the divider has incorrect type (expected QWidget, got {type(divider)})")
        elif isinstance(_divider, QWidget):
            return _divider
        elif _divider is True:
            line = QFrame()
            _direction = self._get_direction()
            line.setFrameShape(QFrame.HLine if _direction in ["column", "column-reverse"] else QFrame.VLine)
            line.setFrameShadow(QFrame.Sunken)
            return line
        return None

    def _parse_alignment(self, alignItems):
        """Parse alignment to Qt alignment based on direction."""
        _direction = self._get_direction()
        _alignItems = self._get_alignItems()
        align_map = {
            "flex-start": Qt.AlignLeft if _direction in ["row", "row-reverse"] else Qt.AlignTop,
            "center": Qt.AlignCenter,
            "flex-end": Qt.AlignRight if _direction in ["row", "row-reverse"] else Qt.AlignBottom,
            "stretch": Qt.AlignJustify,
            "baseline": Qt.AlignTop if _direction in ["row", "row-reverse"] else Qt.AlignLeft
        }
        return align_map.get(_alignItems, Qt.AlignCenter)



    def _resolve_breakpoint_value(self, config_dict, current_bp):
        if not isinstance(config_dict, dict):
            return config_dict

        breakpoints = ["xs", "sm", "md", "lg", "xl"]
        if current_bp not in breakpoints:
            current_bp = "xs"

        bp_index = breakpoints.index(current_bp)
        if current_bp in config_dict:
            return config_dict[current_bp]

        for i in range(bp_index - 1, -1, -1):
            bp = breakpoints[i]
            if bp in config_dict:
                return config_dict[bp]

        return None

    def _get_breakpoint(self):
        width = QApplication.instance().mainWindow.width()
        if width >= 1536:
            return "xl"
        elif width >= 1200:
            return "lg"
        elif width >= 900:
            return "md"
        elif width >= 600:
            return "sm"
        else:
            return "xs"


    def _schedule_repaint(self):
        """
        phải được gọi mỗi khi có thay đổi sx
        """
        if self._repaint_scheduled:
            return
        self._repaint_scheduled = True

        def _go():
            self._repaint_scheduled = False
            self.update()

        QTimer.singleShot(0, _go)
    
    def showEvent(self, event):
        if self._asynRenderQss:
            threading.Thread(target=self._renderStylesheet, args=(), daemon=True).start()
        return super().showEvent(event)
    
    def resizeEvent(self, event):
        PyWidgetBase.resizeEvent(self, event)
        # return super().resizeEvent(event)
        super().resizeEvent(event)
        
        if hasattr(self, "_decor"):
            self._decor.invalidate_cache()
            self._schedule_repaint()

        
    def paintEvent(self, arg__1):
        """Handle paint events."""
        PyWidgetBase.paintEvent(self, arg__1)
        # return super().paintEvent(arg__1)
        super().paintEvent(arg__1)
        
        if hasattr(self, "_decor"):
            p = QPainter(self)
            self._decor.paint(p, self.size(), self.devicePixelRatioF())
            p.end()


# ///////////////////////////////////////////////////////////////
#
# BY: WANDERSON M.PIMENTA
# PROJECT MADE WITH: Qt Designer and ..site_packages.qtcompat
# V: 1.0.0
#
# This project can be used freely for all uses, as long as they maintain the
# respective credits only in the Python scripts, any information in the visual
# interface (GUI) can be modified without any implication.
#
# There are limitations on Qt licenses if you want to use your products
# commercially, I recommend reading them on the official website:
# https://doc.qt.io/qtforpython/licenses.html
#
# ///////////////////////////////////////////////////////////////

# IMPORT PACKAGES AND MODULES
# ///////////////////////////////////////////////////////////////
import json
import os

# IMPORT SETTINGS
# ///////////////////////////////////////////////////////////////
from .json_settings import Settings

# APP THEMES
# ///////////////////////////////////////////////////////////////
class Themes(object):
    # LOAD SETTINGS
    # ///////////////////////////////////////////////////////////////
    setup_settings = Settings()
    _settings = setup_settings.items

    # APP PATH
    # ///////////////////////////////////////////////////////////////
    json_file = f"src/themes/{_settings['theme_name']}.json"
    app_path = os.path.abspath(os.getcwd())
    settings_path = os.path.normpath(os.path.join(app_path, json_file))
    if not os.path.isfile(settings_path):
        print(f"WARNING: \"gui/themes/{_settings['theme_name']}.json\" not found! check in the folder {settings_path}")

    # INIT SETTINGS
    # ///////////////////////////////////////////////////////////////
    def __init__(self):
        super(Themes, self).__init__()

        # DICTIONARY WITH SETTINGS
        self.items = {}

        # DESERIALIZE
        self.deserialize()

    # SERIALIZE JSON
    # ///////////////////////////////////////////////////////////////
    def serialize(self):
        # WRITE JSON FILE
        with open(self.settings_path, "w", encoding='utf-8') as write:
            json.dump(self.items, write, indent=4)

    # DESERIALIZE JSON
    # ///////////////////////////////////////////////////////////////
    def deserialize(self):
        self.items = {
            "app_name": "PyOneDark - Modern GUI",
            "version" : "v1.0.0",
            "copyright" : "By: Wanderson M. Pimenta",
            "year" : 2021,
            "theme_name" : "bright_theme",
            "custom_title_bar": True,
            "startup_size": [
                1400,
                720
            ],
            "minimum_size": [
                960,
                540
            ],
            "lef_menu_size" : {
                "minimum" : 50,
                "maximum" : 240
            },
            "left_menu_content_margins" : 3,
            "left_column_size" : {
                "minimum" : 0,
                "maximum" : 240
            },
            "right_column_size" : {
                "minimum" : 0,
                "maximum" : 240
            },
            "time_animation" : 500,
            "font" : {
                "family" : "Segoe UI",
                "title_size" : 10,
                "text_size" : 9
            }
        }


# tests for win32gui
import array
import operator
import sys
import unittest

import pywin32_testutil
import pywintypes
import win32api
import win32gui


class TestPyGetString(unittest.TestCase):
    def test_get_string(self):
        # test invalid addresses cause a ValueError rather than crash!
        self.assertRaises(ValueError, win32gui.PyGetString, 0)
        self.assertRaises(ValueError, win32gui.PyGetString, 1)
        self.assertRaises(ValueError, win32gui.PyGetString, 1, 1)


class TestPyGetMemory(unittest.TestCase):
    def test_ob(self):
        # Check the PyGetMemory result and a bytes string can be compared
        test_data = b"\0\1\2\3\4\5\6"
        c = array.array("b", test_data)
        addr, buflen = c.buffer_info()
        got = win32gui.PyGetMemory(addr, buflen)
        self.assertEqual(len(got), len(test_data))
        self.assertEqual(bytes(got), test_data)

    def test_memory_index(self):
        # Check we can index into the buffer object returned by PyGetMemory
        test_data = b"\0\1\2\3\4\5\6"
        c = array.array("b", test_data)
        addr, buflen = c.buffer_info()
        got = win32gui.PyGetMemory(addr, buflen)
        self.assertEqual(got[0], 0)

    def test_memory_slice(self):
        # Check we can slice the buffer object returned by PyGetMemory
        test_data = b"\0\1\2\3\4\5\6"
        c = array.array("b", test_data)
        addr, buflen = c.buffer_info()
        got = win32gui.PyGetMemory(addr, buflen)
        self.assertEqual(list(got[0:3]), [0, 1, 2])

    def test_real_view(self):
        # Do the PyGetMemory, then change the original memory, then ensure
        # the initial object we fetched sees the new value.
        test_data = b"\0\1\2\3\4\5\6"
        c = array.array("b", test_data)
        addr, buflen = c.buffer_info()
        got = win32gui.PyGetMemory(addr, buflen)
        self.assertEqual(got[0], 0)
        c[0] = 1
        self.assertEqual(got[0], 1)

    def test_memory_not_writable(self):
        # Check the buffer object fetched by PyGetMemory isn't writable.
        test_data = b"\0\1\2\3\4\5\6"
        c = array.array("b", test_data)
        addr, buflen = c.buffer_info()
        got = win32gui.PyGetMemory(addr, buflen)
        self.assertRaises(TypeError, operator.setitem, got, 0, 1)


class TestEnumWindowsFamily(unittest.TestCase):
    @classmethod
    def enum_callback_sle(cls, handle, data):
        win32api.SetLastError(1)
        return data

    @classmethod
    def enum_callback_exc(cls, handle, data):
        raise ValueError

    @classmethod
    def enum_callback(cls, handle, data):
        return data

    def setUp(self):
        self.default_data_set = (None, -1, 0, 1, True, False)
        self.type_data_set = ("", (), {})

    def test_enumwindows(self):
        win32api.SetLastError(0)
        for data in (0, False):
            self.assertRaises(
                pywintypes.error, win32gui.EnumWindows, self.enum_callback_sle, data
            )
        for data in (None, 1, True):
            self.assertIsNone(win32gui.EnumWindows(self.enum_callback_sle, data))
        win32api.SetLastError(0)
        for data in self.default_data_set:
            self.assertIsNone(win32gui.EnumWindows(self.enum_callback, data))
        for data in self.default_data_set:
            self.assertRaises(
                ValueError, win32gui.EnumWindows, self.enum_callback_exc, data
            )
        for func in (
            self.enum_callback,
            self.enum_callback_sle,
        ):
            for data in self.type_data_set:
                self.assertRaises(TypeError, win32gui.EnumWindows, func, data)
        if sys.version_info >= (3, 10):
            for func in (
                self.enum_callback,
                self.enum_callback_sle,
            ):
                self.assertRaises(
                    TypeError, win32gui.EnumWindows, func, self.enum_callback, 2.718282
                )

    def test_enumchildwindows(self):
        win32api.SetLastError(0)
        for data in self.default_data_set:
            self.assertIsNone(win32gui.EnumChildWindows(None, self.enum_callback, data))
        for data in self.default_data_set:
            self.assertIsNone(
                win32gui.EnumChildWindows(None, self.enum_callback_sle, data)
            )
        win32api.SetLastError(0)
        for data in self.default_data_set:
            self.assertRaises(
                ValueError,
                win32gui.EnumChildWindows,
                None,
                self.enum_callback_exc,
                data,
            )
        for data in self.type_data_set:
            for func in (
                self.enum_callback,
                self.enum_callback_sle,
            ):
                self.assertRaises(
                    TypeError, win32gui.EnumChildWindows, None, func, data
                )
        if sys.version_info >= (3, 10):
            for func in (
                self.enum_callback,
                self.enum_callback_sle,
            ):
                self.assertRaises(
                    TypeError,
                    win32gui.EnumChildWindows,
                    None,
                    func,
                    self.enum_callback,
                    2.718282,
                )

    def test_enumdesktopwindows(self):
        win32api.SetLastError(0)
        desktop = None
        for data in (0, False):
            self.assertRaises(
                pywintypes.error,
                win32gui.EnumDesktopWindows,
                desktop,
                self.enum_callback_sle,
                data,
            )
        for data in (None, 1, True):
            self.assertIsNone(
                win32gui.EnumDesktopWindows(desktop, self.enum_callback_sle, data)
            )
        win32api.SetLastError(0)
        for data in self.default_data_set:
            self.assertIsNone(
                win32gui.EnumDesktopWindows(desktop, self.enum_callback, data)
            )
        for data in self.default_data_set:
            self.assertRaises(
                ValueError,
                win32gui.EnumDesktopWindows,
                desktop,
                self.enum_callback_exc,
                data,
            )
        desktops = (0, None)
        for desktop in desktops:
            for data in self.default_data_set:
                self.assertRaises(
                    ValueError,
                    win32gui.EnumDesktopWindows,
                    desktop,
                    self.enum_callback_exc,
                    data,
                )
        for func in (
            self.enum_callback,
            self.enum_callback_sle,
        ):
            for desktop in desktops:
                for data in self.type_data_set:
                    self.assertRaises(
                        TypeError, win32gui.EnumDesktopWindows, 0, func, data
                    )
        if sys.version_info >= (3, 10):
            for func in (
                self.enum_callback,
                self.enum_callback_sle,
            ):
                for desktop in desktops:
                    self.assertRaises(
                        TypeError, win32gui.EnumDesktopWindows, 0, func, 2.718282
                    )


class TestWindowProperties(unittest.TestCase):
    def setUp(self):
        self.class_functions = (
            win32gui.GetClassName,
            win32gui.RealGetWindowClass,
        )

    def test_classname(self):
        for func in self.class_functions:
            self.assertRaises(pywintypes.error, func, 0)
        wnd = win32gui.GetDesktopWindow()
        for func in self.class_functions:
            self.assertTrue(func(wnd))


if __name__ == "__main__":
    unittest.main()



from typing import TYPE_CHECKING, Callable
from PySide6.QtWidgets import QFrame, QHBoxLayout, QWidget
from PySide6.QtCore import Qt

from ...typography import Typography
from ...avatar import Avatar
from ...label import Label

if TYPE_CHECKING:
    from .table import TableWidget


class TableCell(QFrame):
    def __init__(
                self,
                key: str = None,
                padding: str = "checkbox",
                align: str = "left",
                children: object = None,
                colSpan: int = None,
                onClick: Callable = None,
                sx: str = None,
                text: str = None,
                ):
        super().__init__()

        self._align = align
        self._children = children
        self._text = text
        self._sx = sx

        self._indexRow = None

        # self.setAttribute(Qt.WA_TransparentForMouseEvents)
        self.setLayout(QHBoxLayout())
        self.layout().setContentsMargins(0,0,0,0)

        if self._text:
                self.layout().addWidget(Typography(text=self._text, variant="body2", align=self._align))

        elif self._children:
            if isinstance(self._children, list):
                for widget in self._children:
                    if widget is not None:
                        self.layout().addWidget(widget)
            elif isinstance(self._children, QWidget):
                self.layout().addWidget(self._children)
            elif isinstance(self._children, str):
                cell_widget = Typography(text=self._children, variant="body2", align=self._align)
                self.layout().addWidget(cell_widget)

    
    def enterEvent(self, event):
        try:
            self.parent().parent()._setHoverRow(self._indexRow)
        except Exception as e:
            pass
        # return super().enterEvent(event)


from ....system.color_manipulator import alpha
from .properties_name import *

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from qtmui.material.styles.create_theme.theme_reducer import ThemeState


COLORS = ['primary', 'secondary', 'info', 'success', 'warning', 'error']

def button_group(_theme):
    theme: ThemeState = _theme
    
    VARIANTS = ["contained", "outlined", "text", "soft"]


        # default_style = {
        #     f"& .{buttonGroupClasses.grouped}": {
        #         '&:not(:last-of-type)': {
        #             **({
        #                 'borderStyle': 'solid',
        #                 **({
        #                     borderColor: alpha(theme.palette.grey._500, 0.32),
        #                 } if inherit_color else {}),
        #                 **({
        #                     'borderWidth': '0px 1px 0px 0px',
        #                 } if horizontal_orientation else {}),
        #                 **({
        #                     'borderWidth': '0px 0px 1px 0px',
        #                 } if vertical_orientation else {}),
        #             } if not outlined_variant else {}),
        #         },
        #     },
        # }

        # color_styles = [
        #     {
        #         f"& .{buttonGroupClasses.grouped}": {
        #             '&:not(:last-of-type)': {
        #                 **({
        #                     **({
        #                         borderColor: alpha(getattr(theme.palette, color).dark, 0.48),
        #                     } if contained_variant else {}),
        #                     **({
        #                         borderColor: alpha(getattr(theme.palette, color).main, 0.48),
        #                     } if text_variant else {}),
        #                     **({
        #                         borderColor: alpha(getattr(theme.palette, color).dark, 0.24),
        #                     } if soft_variant else {}),
        #                 } if owner_state.get('color') == color else {}),
        #             },
        #         },
        #     } for color in COLORS
        # ]

        # disabled_state = {
        #     f"& .{buttonGroupClasses.grouped}": {
        #         f"&.{buttonGroupClasses.disabled}": {
        #             '&:not(:last-of-type)': {
        #                 borderColor: theme.palette.action.disabledBackground,
        #             },
        #         },
        #     },
        # }


    return {
        'MuiButtonGroup': {
            'styles': {
                'root': {
                        f"{variant}": {
                            "first": {
                                borderRadius: '0px',
                                "orientation": {
                                    "vertical": {
                                        borderTopLeftRadius: "4px",
                                        borderTopRightRadius: "4px",
                                        borderWidth: "0px 0px 1px 0px"
                                    },
                                    "horizontal": {
                                        borderTopLeftRadius: "4px",
                                        borderBottomLeftRadius: "4px",
                                        borderWidth: "0px 1px 0px 0px"
                                    },
                                },
                                "borderColorDefault": theme.palette.grey._500,
                                "borderColorRender": lambda color, _variant=variant: 
                                getattr(theme.palette, color).dark if _variant == "contained" 
                                else  alpha(getattr(theme.palette, color).main, 0.48) if _variant == "text" 
                                else  alpha(getattr(theme.palette, color).main, 0.24) if _variant == "soft" 
                                else None
                            },
                            "middle": {
                                borderRadius: '0px',
                                "orientation": {
                                    "vertical": {
                                        borderWidth: "0px 0px 1px 0px"
                                    },
                                    "horizontal": {
                                        borderWidth: "0px 1px 0px 0px"
                                    },
                                },
                            },
                            "last": {
                                borderRadius: '0px',
                                borderWidth: "0px",
                                "orientation": {
                                    "vertical": {
                                        borderBottomLeftRadius: "4px",
                                        borderBottomRightRadius: "4px",
                                    },
                                    "horizontal": {
                                        borderTopRightRadius: "4px",
                                        borderBottomRightRadius: "4px",
                                    },
                                },
                            },
                        }
                        for variant in VARIANTS
                    }
            },
        },
    }



from PySide6.QtWidgets import QWidget
from PySide6.QtGui import QPixmap, QPainter, QBrush
from PySide6.QtCore import Qt

def widget_to_pixmap(widget):
    """Chuyển đổi một QWidget thành QPixmap."""
    # Tạo một đối tượng QPixmap với kích thước của widget
    pixmap = QPixmap(widget.size())
    
    # Bắt đầu vẽ nội dung của widget lên pixmap
    widget.render(pixmap)
    
    return pixmap


class PixmapWidget(QWidget):
    def __init__(self, widget):
        super().__init__()

        self._widget = widget
        self._width = widget.width()
        self._height = widget.height()


        self.setFixedSize(widget.size())

        

    def paintEvent(self, event):
        # Chuyển đổi widget thành QPixmap
        pixmap = widget_to_pixmap(self._widget)
        # Lưu QPixmap ra file ảnh PNG để kiểm tra
        # pixmap.save("widget_capture.png", "PNG")
        scaled_pixmap = pixmap.scaled(self.size(), Qt.AspectRatioMode.KeepAspectRatioByExpanding, Qt.TransformationMode.SmoothTransformation)

        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)  

        brush = QBrush()
        brush.setStyle(Qt.Dense2Pattern)
        brush.setTexture(pixmap)
        painter.setPen(Qt.NoPen)
        painter.setBrush(brush)

        painter.drawRoundedRect(self.rect(), self._width, self._height)

        super().paintEvent(event)


# if __name__ == "__main__":
#     app = QApplication([])

#     # Tạo một widget ví dụ
#     widget = PixmapWidget()
#     widget.show()
#     app.exec()



from typing import Tuple, Union, Dict, List
from urllib.parse import parse_qs

class RouterParse():
    def __init__(self, list_router: List[dict]):
        self.map_all_route:list = list_router

    def paire_route(self,route:str="dashboard/user"):
        paire_routes = []
        if "/" in route:
            routes = route.split("/")
            for i in routes:
                if i == "":
                    continue
                paire_routes.append(i)
        else:
            paire_routes = [route]
        return paire_routes
    
    def get_dict_route(self,main_path:str="dashboard"):
        for route in self.map_all_route:
            path = route.get("path")
            if path == main_path:
                return route
        return {}
    
    def get_index_true(self,children:list=[]):
        for child in children:
            if child.get("index") == True:
                return child
        return {}
    
    def get_current_route(self,chilren:list=[], path:str=""):
        for route in chilren:
            if route.get("path") == path:
                return route
        return {}
    
    def extract_param_and_clean_path(self, path: str):
        """
        Hàm nhận một đường dẫn có chứa segment với dấu ':' và trả về:
        - Giá trị của segment đó (sau khi loại bỏ dấu ':')
        - Đường dẫn mới, loại bỏ segment chứa dấu ':'
        
        Ví dụ:
        Input: "/dashboard/user/:5/edit"
        Output: ("5", "/dashboard/user/edit")
        
        Input: "/dashboard/user/edit/:5"
        Output: ("5", "/dashboard/user/edit")
        """
        param = None

        if path.startswith("push//:"):
            path = path.split(':')
            new_path = path[1]
            param = path[0]
            path_and_params = new_path.split('?')
            new_path = path_and_params[0]
            param = path_and_params[1]
            # param=params.split("&&")
            return param, new_path
        
        elif path.startswith("replace//:"):
            path = path.split(':')
            new_path = path[1]
            param = path[0]
            return param, new_path

        elif path.find("/?") != -1:
            path = path.split('/?')
            new_path = path[0]
            param = path[1]
            return param, new_path

        segments = path.split('/')
        
        new_segments = []
        
        for seg in segments:
            if ':' in seg:
                # Lấy phần sau dấu ':' (có thể là toàn bộ chuỗi sau khi loại bỏ dấu ':')
                param = seg.replace(":", "")
            else:
                new_segments.append(seg)
        
        # Ghép lại các segment thành đường dẫn mới
        new_path = "/".join(new_segments)
        # Nếu đường dẫn ban đầu bắt đầu bằng '/', đảm bảo kết quả cũng bắt đầu bằng '/'
        if path.startswith("/") and not new_path.startswith("/"):
            new_path = "/" + new_path

        return param, new_path

    def extract_param_and_clean_path_new(self, path: str) -> Tuple[Union[str, Dict, None], str]:
        """
        Hàm nhận một đường dẫn và trả về:
        - Giá trị tham số (chuỗi nếu là :param, dictionary nếu là query string, hoặc None)
        - Đường dẫn mới, loại bỏ segment chứa :param hoặc query string
        
        Ví dụ:
        Input: "/dashboard/user/:5/edit"
        Output: ("5", "/dashboard/user/edit")
        
        Input: "/dashboard/user/edit?userName=6&password=4"
        Output: ({"userName": "6", "password": "4"}, "/dashboard/user/edit")
        
        Input: "/dashboard/user/edit"
        Output: (None, "/dashboard/user/edit")
        """
        param = None
        new_path = path

        # Xử lý query string (?key=value)
        if '?' in path:
            path_parts = path.split('?', 1)
            new_path = path_parts[0]  # Phần path trước query string
            query_string = path_parts[1] if len(path_parts) > 1 else ""
            
            # # Phân tích query string thành dictionary
            # if query_string:
            #     # parse_qs trả về dict với value là list (vì query string có thể lặp key)
            #     param_dict = parse_qs(query_string)
            #     # Chuyển value từ list sang single value nếu chỉ có 1 phần tử
            #     param = {key: value[0] if len(value) == 1 else value for key, value in param_dict.items()}
            # else:
            #     param = {}
            
            return query_string, new_path

        # Xử lý các trường hợp push//: và replace//:
        if path.startswith("push//:") or path.startswith("replace//:"):
            path_parts = path.split(':', 1)
            new_path = path_parts[1] if len(path_parts) > 1 else path
            param = path_parts[0]
            
            # Xử lý query string trong push//: hoặc replace//:
            if '?' in new_path:
                path_and_params = new_path.split('?', 1)
                new_path = path_and_params[0]
                query_string = path_and_params[1] if len(path_and_params) > 1 else ""
                if query_string:
                    param_dict = parse_qs(query_string)
                    param = {key: value[0] if len(value) == 1 else value for key, value in param_dict.items()}
            
            return param, new_path

        # Xử lý trường hợp :param trong path
        segments = path.split('/')
        new_segments = []
        
        for seg in segments:
            if ':' in seg:
                param = seg.replace(":", "")
            else:
                new_segments.append(seg)
        
        # Ghép lại các segment thành đường dẫn mới
        new_path = "/".join(new_segments)
        # Đảm bảo đường dẫn bắt đầu bằng '/' nếu đầu vào có
        if path.startswith("/") and not new_path.startswith("/"):
            new_path = "/" + new_path

        return param, new_path

    def get_route(self, route:str="//dashboard///user///profile///", default=None):
        param = None
        if route.find(":") != -1:
            param, route = self.extract_param_and_clean_path(route)
        paths = self.paire_route(route)
        main_path = paths[0]
        current:dict = self.get_dict_route(main_path)
        layout = None
        page = None
        leng = len(paths)
        for i in range(leng):
            path = paths[i]
            _path = current.get("path")
            _type = current.get("type")
            element = current.get("element")
            children = current.get("children")
            index = current.get("index")
            
            if _type == "layout":
                layout = element
            
            if i == leng-1:
                if children:
                    index_true = self.get_index_true(children)
                    # print("index_true", index_true)
                    page = index_true.get("element")
                    return {"layout":layout, "page":page, "param": param}
                page = current.get("element")
                if page == layout:
                    page = None
                return {"layout":layout, "page":page, "param": param}
            else:
                if children:
                    next_path = paths[i+1]
                    current = self.get_current_route(children,next_path)
                    continue
                return {"layout":layout, "page":page, "param": param}
            
# route = RouterParse(list_router=list_rourte)
# # kq = route.get_route("//dashboard---///a--pp//gfgd///")
# kq = route.get_route("dashboard/app")
# kq = route.get_route("dashboard/user")
# print(kq)




from typing import Callable, Optional, Union
from PySide6.QtWidgets import QHBoxLayout, QFrame, QSpacerItem, QSizePolicy

from qtmui.hooks import State

from ..typography.typography import Typography
from ..button import IconButton

class DialogHeader(QFrame):
    def __init__(
            self,
            parent=None,
            title: Optional[Union[State, str, Callable]] = "Information",
            align: str = "right"
        ):
        super().__init__(parent)
        self.setLayout(QHBoxLayout())
        
        if align == "left":
            self.layout().addWidget(IconButton(onClick=parent.hide_dialog, icon=":/round/resource_qtmui/round/close.svg", variant="soft"))
            self.layout().addItem(QSpacerItem(448, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))
        else:
            self.layout().addWidget(Typography(text=title, variant="h3"))
            self.layout().addItem(QSpacerItem(448, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))
            self.layout().addWidget(IconButton(onClick=parent.hide_dialog, icon=":/round/resource_qtmui/round/close.svg", variant="soft"))


from typing import Dict, TYPE_CHECKING
if TYPE_CHECKING:
    from qtmui.material.styles.create_theme.theme_reducer import ThemeState

from ....system.color_manipulator import alpha
from .properties_name import *

def switch(_theme) -> Dict:
    theme: ThemeState = _theme
    lightMode = theme.palette.mode == 'light'
    COLORS = ['primary', 'secondary', 'info', 'success', 'warning', 'error']

    return {
        'PySwitch': {
            'styles': {
                'root': {
                    **{
                        "default": {
                            'indicatorColorChecked': theme.palette.grey._800 if lightMode else theme.palette.common.white,
                            'barColor': theme.palette.grey._500,
                            'handleColor': theme.palette.common.white if lightMode else theme.palette.grey._800,
                            'pulseCheckedColor': theme.palette.grey._700 if lightMode else theme.palette.grey._400,
                            'pulseUncheckedColor': theme.palette.common.white if lightMode else theme.palette.grey._800,
                            'checkedColor': theme.palette.grey._800 if lightMode else theme.palette.common.white,
                        },
                        "inherit": {
                            'indicatorColorChecked': theme.palette.grey._800 if lightMode else theme.palette.common.white,
                            'barColor': theme.palette.grey._500,
                            'handleColor': theme.palette.grey._800 if lightMode else theme.palette.common.white,
                            'pulseCheckedColor': theme.palette.grey._700 if lightMode else theme.palette.grey._400,
                            'pulseUncheckedColor': theme.palette.common.white if lightMode else theme.palette.grey._800,
                            'checkedColor': theme.palette.grey._800 if lightMode else theme.palette.common.white
                        }
                    },
                    **{
                        f"{_color}": {
                            'indicatorColorChecked': getattr(theme.palette, _color).main,
                            'barColor': theme.palette.grey._500,
                            'handleColor': theme.palette.common.white,
                            'pulseCheckedColor': getattr(theme.palette, _color).light if lightMode else getattr(theme.palette, _color).dark,
                            'pulseUncheckedColor': getattr(theme.palette, _color).main,
                            'checkedColor': getattr(theme.palette, _color).main,
                        }
                        for _color in COLORS
                    }
                }
            }
        },
    }



# qtmui/utils/validate_params.py
from typing import Any, Optional, Union, Dict, List, get_args, get_origin

from qtmui.errors import PyMuiValidationError
from qtmui.hooks import State
from PySide6.QtWidgets import QWidget

# Hàm kiểm tra số nguyên dương (positive integer)
def validate_positive_integer(param_name: str, value: Optional[Union[int, State]], allow_none: bool = True, **kwargs) -> Optional[str]:
    """
    Kiểm tra tham số có phải là số nguyên dương hoặc State chứa số nguyên dương.
    :param param_name: Tên của tham số cần kiểm tra (ví dụ: 'columns').
    :param value: Giá trị cần kiểm tra (có thể là int hoặc State chứa int).
    :param allow_none: Cho phép giá trị None hay không (mặc định là True).
    :return: Thông điệp lỗi nếu có, hoặc None nếu hợp lệ.
    """
    if allow_none and value is None:
        return None
    actual_value = value.value if isinstance(value, State) else value
    if isinstance(value, State):
        if not isinstance(actual_value, int) or actual_value <= 0:
            return "Parameter '" + str(param_name) + "' must be a positive integer, but got '" + str(actual_value) + "'"
    elif not isinstance(actual_value, int) or actual_value <= 0:
        return "Parameter '" + str(param_name) + "' must be a positive integer, but got '" + str(actual_value) + "'"
    return None

# Hàm kiểm tra số dương (positive number)
def validate_positive_number(param_name: str, value: Optional[Union[int, float, State]], allow_none: bool = True, **kwargs) -> Optional[str]:
    """
    Kiểm tra tham số có phải là số dương (lớn hơn 0) hoặc State chứa số dương.
    :param param_name: Tên của tham số cần kiểm tra.
    :param value: Giá trị cần kiểm tra (có thể là int, float hoặc State chứa int/float).
    :param allow_none: Cho phép giá trị None hay không (mặc định là True).
    :return: Thông điệp lỗi nếu có, hoặc None nếu hợp lệ.
    """
    if allow_none and value is None:
        return None
    actual_value = value.value if isinstance(value, State) else value
    if isinstance(value, State):
        if not isinstance(actual_value, (int, float)) or actual_value <= 0:
            return "Parameter '" + str(param_name) + "' must be a positive number, but got '" + str(actual_value) + "'"
    elif not isinstance(actual_value, (int, float)) or actual_value <= 0:
        return "Parameter '" + str(param_name) + "' must be a positive number, but got '" + str(actual_value) + "'"
    return None

# Hàm kiểm tra số không âm (non-negative number)
def validate_non_negative_number(param_name: str, value: Optional[Union[int, float, State]], allow_none: bool = True, **kwargs) -> Optional[str]:
    """
    Kiểm tra tham số có phải là số không âm (0 hoặc lớn hơn) hoặc State chứa số không âm.
    :param param_name: Tên của tham số cần kiểm tra (ví dụ: 'spacing').
    :param value: Giá trị cần kiểm tra (có thể là int, float hoặc State chứa int/float).
    :param allow_none: Cho phép giá trị None hay không (mặc định là True).
    :return: Thông điệp lỗi nếu có, hoặc None nếu hợp lệ.
    """
    if allow_none and value is None:
        return None
    actual_value = value.value if isinstance(value, State) else value
    if isinstance(value, State):
        if not isinstance(actual_value, (int, float)) or actual_value < 0:
            return "Parameter '" + str(param_name) + "' must be a non-negative number, but got '" + str(actual_value) + "'"
    elif not isinstance(actual_value, (int, float)) or actual_value < 0:
        return "Parameter '" + str(param_name) + "' must be a non-negative number, but got '" + str(actual_value) + "'"
    return None

# Hàm kiểm tra spacing (khoảng cách)
def validate_spacing(param_name: str, value: Optional[Union[int, Dict, State]], allow_none: bool = True, **kwargs) -> Optional[str]:
    """
    Kiểm tra tham số có phải là số không âm, dict chứa số không âm, hoặc State chứa giá trị như vậy.
    :param param_name: Tên của tham số cần kiểm tra (ví dụ: 'spacing').
    :param value: Giá trị cần kiểm tra (có thể là int, dict hoặc State chứa int/dict).
    :param allow_none: Cho phép giá trị None hay không (mặc định là True).
    :return: Thông điệp lỗi nếu có, hoặc None nếu hợp lệ.
    """
    if allow_none and value is None:
        return None
    actual_value = value.value if isinstance(value, State) else value
    if isinstance(value, State):
        if isinstance(actual_value, dict):
            for key, val in actual_value.items():
                if not isinstance(val, (int, float)) or val < 0:
                    return "Value of parameter '" + str(param_name) + "' for breakpoint '" + str(key) + "' must be a non-negative number, but got '" + str(val) + "'"
        elif not isinstance(actual_value, (int, float)) or actual_value < 0:
            return "Parameter '" + str(param_name) + "' must be a non-negative number or dict containing non-negative numbers, but got '" + str(actual_value) + "'"
    elif isinstance(value, dict):
        for key, val in value.items():
            if not isinstance(val, (int, float)) or val < 0:
                return "Value of parameter '" + str(param_name) + "' for breakpoint '" + str(key) + "' must be a non-negative number, but got '" + str(val) + "'"
    elif not isinstance(actual_value, (int, float)) or actual_value < 0:
        return "Parameter '" + str(param_name) + "' must be a non-negative number or dict containing non-negative numbers, but got '" + str(actual_value) + "'"
    return None

# Hàm kiểm tra sx (style tùy chỉnh)
def validate_sx(param_name: str, value: Optional[Union[State, str, Dict]], allow_none: bool = True, **kwargs) -> Optional[str]:
    """
    Kiểm tra tham số có phải là dict, str, hoặc State chứa một trong các kiểu này.
    :param param_name: Tên của tham số cần kiểm tra (ví dụ: 'sx').
    :param value: Giá trị cần kiểm tra (có thể là dict, str hoặc State chứa kiểu tương ứng).
    :param allow_none: Cho phép giá trị None hay không (mặc định là True).
    :return: Thông điệp lỗi nếu có, hoặc None nếu hợp lệ.
    """
    if allow_none and value is None:
        return None
    actual_value = value.value if isinstance(value, State) else value
    if isinstance(value, State):
        if not isinstance(actual_value, (dict, str)):
            return "Parameter '" + str(param_name) + "' must be a dict, str, or State containing one of these types, but got '" + str(actual_value) + "'"
    elif not isinstance(actual_value, (dict, str)):
        return "Parameter '" + str(param_name) + "' must be a dict, str, or State containing one of these types, but got '" + str(actual_value) + "'"
    return None

# Hàm kiểm tra giá trị boolean
def validate_bool(param_name: str, value: Optional[Union[bool, State]], allow_none: bool = True, **kwargs) -> Optional[str]:
    """
    Kiểm tra tham số có phải là boolean hoặc State chứa boolean.
    :param param_name: Tên của tham số cần kiểm tra (ví dụ: 'container').
    :param value: Giá trị cần kiểm tra (có thể là bool hoặc State chứa bool).
    :param allow_none: Cho phép giá trị None hay không (mặc định là True).
    :return: Thông điệp lỗi nếu có, hoặc None nếu hợp lệ.
    """
    if allow_none and value is None:
        return None
    actual_value = value.value if isinstance(value, State) else value
    if isinstance(value, State):
        if not isinstance(actual_value, bool):
            return "Parameter '" + str(param_name) + "' must be a boolean or State containing a boolean, but got '" + str(actual_value) + "'"
    elif not isinstance(actual_value, bool):
        return "Parameter '" + str(param_name) + "' must be a boolean or State containing a boolean, but got '" + str(actual_value) + "'"
    return None

# Hàm kiểm tra chuỗi
def validate_string(param_name: str, value: Optional[Union[str, State]], allow_none: bool = True, **kwargs) -> Optional[str]:
    """
    Kiểm tra tham số có phải là chuỗi hoặc State chứa chuỗi.
    :param param_name: Tên của tham số cần kiểm tra (ví dụ: 'key').
    :param value: Giá trị cần kiểm tra (có thể là str hoặc State chứa str).
    :param allow_none: Cho phép giá trị None hay không (mặc định là True).
    :return: Thông điệp lỗi nếu có, hoặc None nếu hợp lệ.
    """
    if allow_none and value is None:
        return None
    actual_value = value.value if isinstance(value, State) else value
    if isinstance(value, State):
        if not isinstance(actual_value, str):
            return "Parameter '" + str(param_name) + "' must be a string or State containing a string, but got '" + str(actual_value) + "'"
    elif not isinstance(actual_value, str):
        return "Parameter '" + str(param_name) + "' must be a string or State containing a string, but got '" + str(actual_value) + "'"
    return None

# Hàm kiểm tra giá trị nằm trong danh sách cho phép
def validate_choice(param_name: str, value: Optional[Union[str, State]], valid_values: List[str], allow_none: bool = True, **kwargs) -> Optional[str]:
    """
    Kiểm tra tham số có phải là một trong các giá trị hợp lệ hoặc State chứa giá trị hợp lệ.
    :param param_name: Tên của tham số cần kiểm tra (ví dụ: 'wrap').
    :param value: Giá trị cần kiểm tra (có thể là str hoặc State chứa str).
    :param valid_values: Danh sách các giá trị hợp lệ (ví dụ: ["wrap", "nowrap", "wrap-reverse"]).
    :param allow_none: Cho phép giá trị None hay không (mặc định là True).
    :return: Thông điệp lỗi nếu có, hoặc None nếu hợp lệ.
    """
    if allow_none and value is None:
        return None
    actual_value = value.value if isinstance(value, State) else value
    if isinstance(value, State):
        if not isinstance(actual_value, str) or actual_value not in valid_values:
            return "Parameter '" + str(param_name) + "' must be one of " + str(list(valid_values)) + ", but got '" + str(actual_value) + "'"
    elif not isinstance(actual_value, str) or actual_value not in valid_values:
        return "Parameter '" + str(param_name) + "' must be one of " + str(list(valid_values)) + ", but got '" + str(actual_value) + "'"
    return None

# Hàm kiểm tra dict chứa số nguyên dương
def validate_dict_with_positive_integers(param_name: str, value: Optional[Union[Dict, State]], allow_none: bool = True, **kwargs) -> Optional[str]:
    """
    Kiểm tra tham số có phải là dict hoặc State chứa dict, với các giá trị là số nguyên dương.
    :param param_name: Tên của tham số cần kiểm tra (ví dụ: 'size').
    :param value: Giá trị cần kiểm tra (có thể là dict hoặc State chứa dict).
    :param allow_none: Cho phép giá trị None hay không (mặc định là True).
    :return: Thông điệp lỗi nếu có, hoặc None nếu hợp lệ.
    """
    if allow_none and value is None:
        return None
    actual_value = value.value if isinstance(value, State) else value
    if isinstance(value, State):
        if not isinstance(actual_value, dict):
            return "Parameter '" + str(param_name) + "' must be a dict or State containing a dict, but got '" + str(actual_value) + "'"
        for key, val in actual_value.items():
            if val is not None and (not isinstance(val, int) or val <= 0):
                return "Value of parameter '" + str(param_name) + "' for breakpoint '" + str(key) + "' must be a positive integer, but got '" + str(val) + "'"
    elif not isinstance(actual_value, dict):
        return "Parameter '" + str(param_name) + "' must be a dict or State containing a dict, but got '" + str(actual_value) + "'"
    else:
        for key, val in value.items():
            if val is not None and (not isinstance(val, int) or val <= 0):
                return "Value of parameter '" + str(param_name) + "' for breakpoint '" + str(key) + "' must be a positive integer, but got '" + str(val) + "'"
    return None

# Hàm kiểm tra children (các phần tử con)
def validate_children(param_name: str, value: Optional[Union[str, list, QWidget, State]], allow_none: bool = True, **kwargs) -> Optional[str]:
    """
    Kiểm tra tham số có phải là str, list, QWidget, hoặc State chứa một trong các kiểu này.
    Nếu là list, các phần tử phải là QWidget.
    :param param_name: Tên của tham số cần kiểm tra (ví dụ: 'children').
    :param value: Giá trị cần kiểm tra (có thể là str, list, QWidget hoặc State chứa kiểu tương ứng).
    :param allow_none: Cho phép giá trị None hay không (mặc định là True).
    :return: Thông điệp lỗi nếu có, hoặc None nếu hợp lệ.
    """
    if allow_none and value is None:
        return None
    actual_value = value.value if isinstance(value, State) else value
    if isinstance(value, State):
        if not isinstance(actual_value, (str, list)):
            return "Parameter '" + str(param_name) + "' must be a str, list, or State containing one of these types, but got '" + str(actual_value) + "'"
        if isinstance(actual_value, list):
            for child in actual_value:
                if child is not None and not isinstance(child, QWidget):
                    return "Each element in list '" + str(param_name) + "' must be a QWidget, but got '" + str(child) + "'"
    elif not isinstance(actual_value, (str, list, QWidget)):
        return "Parameter '" + str(param_name) + "' must be a str, list, QWidget, or State containing one of these types, but got '" + str(actual_value) + "'"
    if isinstance(value, list):
        for child in value:
            if child is not None and not isinstance(child, QWidget):
                return "Each element in list '" + str(param_name) + "' must be a QWidget, but got '" + str(child) + "'"
    return None

# Hàm kiểm tra divider (dùng trong Stack)
def validate_divider(param_name: str, value: Optional[Union[QWidget, bool, State]], allow_none: bool = True, **kwargs) -> Optional[str]:
    """
    Kiểm tra tham số có phải là QWidget, bool, hoặc State chứa một trong các kiểu này.
    :param param_name: Tên của tham số cần kiểm tra (ví dụ: 'divider').
    :param value: Giá trị cần kiểm tra (có thể là QWidget, bool hoặc State chứa kiểu tương ứng).
    :param allow_none: Cho phép giá trị None hay không (mặc định là True).
    :return: Thông điệp lỗi nếu có, hoặc None nếu hợp lệ.
    """
    if allow_none and value is None:
        return None
    actual_value = value.value if isinstance(value, State) else value
    if isinstance(value, State):
        if not isinstance(actual_value, (QWidget, bool)):
            return "Parameter '" + str(param_name) + "' must be a QWidget, bool, or State containing one of these types, but got '" + str(actual_value) + "'"
    elif not isinstance(actual_value, (QWidget, bool)):
        return "Parameter '" + str(param_name) + "' must be a QWidget, bool, or State containing one of these types, but got '" + str(actual_value) + "'"
    return None

# Decorator để áp dụng validation cho tham số
def validate_param(validator_func, param_name, allow_none=True, **kwargs):
    """
    Decorator để kiểm tra tham số trước khi gán giá trị.
    :param validator_func: Hàm kiểm tra được sử dụng (ví dụ: validate_spacing).
    :param param_name: Tên của tham số cần kiểm tra (ví dụ: 'spacing').
    :param allow_none: Cho phép giá trị None hay không (mặc định là True).
    :param kwargs: Các tham số bổ sung để truyền vào hàm kiểm tra (ví dụ: valid_values).
    """
    def decorator(func):
        def wrapper(self, value):
            # Tạo một dict chứa các tham số cần thiết cho validator_func
            validator_kwargs = kwargs.copy()
            validator_kwargs["allow_none"] = allow_none  # Truyền allow_none vào validator_func
            actual_value = value.value if isinstance(value, State) else value
            error_message = validator_func(param_name=param_name, value=value, **validator_kwargs)
            if error_message:
                error_type = "INVALID_VALUE" if "must be one of" in error_message else "INVALID_TYPE"
                expected_types = None
                expected_values = validator_kwargs.get("valid_values", None)
                if error_type == "INVALID_TYPE":
                    if validator_func == validate_sx:
                        expected_types = [dict, str]
                    elif validator_func == validate_bool:
                        expected_types = bool
                    elif validator_func == validate_string:
                        expected_types = str
                    elif validator_func == validate_positive_integer:
                        expected_types = int
                    elif validator_func == validate_positive_number:
                        expected_types = [int, float]
                    elif validator_func == validate_non_negative_number:
                        expected_types = [int, float]
                    elif validator_func == validate_spacing:
                        expected_types = [int, float, dict]
                    elif validator_func == validate_dict_with_positive_integers:
                        expected_types = dict
                    elif validator_func == validate_children:
                        expected_types = [str, list, QWidget]
                    elif validator_func == validate_divider:
                        expected_types = [QWidget, bool]
                # raise ValidationError(
                #     component_name=self.__class__.__name__,
                #     param_name=param_name,
                #     value=actual_value,
                #     message=error_message,
                #     expected_types=expected_types,
                #     expected_values=expected_values,
                #     error_type=error_type
                # )
            return func(self, value)
        return wrapper
    return decorator

def _validate_param(file_path: str, param_name: str, supported_signatures: Union[type, List[type], str], valid_values: Optional[List[Any]] = None, validator: Optional[callable] = None):
    """Decorator để kiểm tra tham số trước khi gán giá trị."""
    def decorator(func):
        def wrapper(self, value):
            # Xử lý giá trị thực tế nếu là State
            actual_value = value.value if isinstance(value, State) else value

            # Xử lý supported_signatures dạng Union
            if get_origin(supported_signatures) is Union:
                supported_types = get_args(supported_signatures)
            else:
                supported_types = supported_signatures if isinstance(supported_signatures, (list, tuple)) else [supported_signatures]

            # Kiểm tra kiểu dữ liệu
            if not isinstance(actual_value, tuple(supported_types)):
                error = PyMuiValidationError(
                    file_path=file_path,
                    param_name=param_name,
                    param_value=actual_value,
                    expected_types=supported_signatures,
                    expected_values=valid_values,
                    error_type="INVALID_TYPE"
                )
                raise error

            # Kiểm tra giá trị hợp lệ nếu có
            if valid_values is not None and not isinstance(value, State):
                if actual_value not in valid_values:
                    error = PyMuiValidationError(
                        file_path=file_path,
                        param_name=param_name,
                        param_value=actual_value,
                        expected_types=supported_signatures,
                        expected_values=valid_values,
                        error_type="INVALID_VALUE"
                    )
                    raise error

            # Nếu là State, kiểm tra giá trị của State
            if isinstance(value, State):
                if not isinstance(actual_value, tuple(supported_types)):
                    error = PyMuiValidationError(
                        file_path=file_path,
                        param_name=param_name,
                        param_value=actual_value,
                        expected_types=supported_signatures,
                        expected_values=valid_values,
                        error_type="INVALID_TYPE"
                    )
                    raise error
                if valid_values is not None and actual_value not in valid_values:
                    error = PyMuiValidationError(
                        file_path=file_path,
                        param_name=param_name,
                        param_value=actual_value,
                        expected_types=supported_signatures,
                        expected_values=valid_values,
                        error_type="INVALID_VALUE"
                    )
                    raise error

            # Kiểm tra điều kiện tùy chỉnh (validator) nếu có
            if validator is not None:
                check_value = value.value if isinstance(value, State) else value
                if not validator(check_value):
                    error = PyMuiValidationError(
                        file_path=file_path,
                        param_name=param_name,
                        param_value=check_value,
                        expected_types=supported_signatures,
                        error_type="INVALID_CONDITION",
                        message="Parameter '" + str(param_name) + "' does not satisfy the validation condition: " + str(check_value)
                    )
                    raise error

            return func(self, value)
        return wrapper
    return decorator